#pragma once

#include <string>
#include <cgv/base/linker.h>
#include <cgv/base/base.h>
#include <cgv/base/lib_begin.h>

namespace cgv {
	namespace base {

/// factory classes that allow to create objects derived from named_object
struct CGV_API abst_factory
{
	base_ptr singelton;
	bool is_singelton;
	abst_factory(bool _singelton = false);
	virtual base_ptr create_object() = 0;
};

/// interfaces that allows to listen to registration events
struct CGV_API abst_registration_listener
{
	virtual void register_object(base_ptr object) = 0;
	virtual void register_factory(abst_factory* factory, const char* item_text, char shortcut) = 0;
};

/// if registration is disable, all registration events are stored and sent at the momement when registration is enabled again. This feature is used when loading dlls
extern void CGV_API disable_registration();

/// enable registration and send all registration events that where emitted during disabled registration
extern void CGV_API enable_registration();

/// check whether registration is enabled
extern bool CGV_API is_registration_enabled();

/// add a listener for factory registration events
extern void CGV_API add_registration_listener(abst_registration_listener* listener);

/// register a factory and send event to all current registration listeners
extern void CGV_API register_factory(abst_factory* factory, const char* item_text, char shortcut);

/// register an object and send event to all current registration listeners
extern void CGV_API register_object(base_ptr object);

/// load a plugin and return a handle to the plugin, or 0 if loading was not successful
extern CGV_API void* load_plugin(const std::string& file_name);

/// convenience class to construct a factory for a given class type
template <class T>
struct factory : public abst_factory
{
	factory(bool is_singleton = false) : abst_factory(is_singleton) {}
	inline base_ptr create_object() { return base_ptr(new T()); }
};

/// convenience class to construct a factory for a class type with a constructor with one argument
template <class T, typename constructor_argument_type>
struct factory_1 : public abst_factory
{
	constructor_argument_type constructor_argument;
	inline factory_1(constructor_argument_type _constructor_argument, bool is_singleton = false) : abst_factory(is_singleton), constructor_argument(_constructor_argument) {}
	inline base_ptr create_object() { return base_ptr(new T(constructor_argument)); }
};

/// convenience class to register a factory of the given class type
template <class T>
struct factory_registration : public force_link
{
	factory_registration(const char* item_text, char shortcut, bool is_singleton = false);
};

/// convenience class to register a factory of the given class type with one constructor argument
template <class T, typename constructor_argument_type>
struct factory_registration_1 : public force_link
{
	factory_registration_1(const char* item_text, char shortcut, constructor_argument_type _constructor_argument, bool is_singleton = false);
};


/// convenience class to register an object of the given class type
template <class T>
struct object_registration : public force_link
{
	/// you must pass something to avoid discarding of code
	object_registration(const char* item_text);
};

/// convenience class to register an object of the given class type with one constructor argument
template <class T, typename construct_arg_type>
struct object_registration_1 : public force_link
{
	/// you must pass something to avoid discarding of code
	object_registration_1(construct_arg_type arg);
};


template <class T>
inline factory_registration<T>::factory_registration(const char* item_text, char shortcut, bool is_singleton)
{
	register_factory(new factory<T>(is_singleton), item_text, shortcut);
}

template <class T, typename constructor_argument_type>
inline factory_registration_1<T, constructor_argument_type>::factory_registration_1(
	const char* item_text, char shortcut, constructor_argument_type _constructor_argument, bool is_singleton)
{
	register_factory(new factory_1<T, constructor_argument_type>(_constructor_argument, is_singleton), item_text, shortcut);
}

template <class T>
object_registration<T>::object_registration(const char* item_text)
{
	register_object(base_ptr(new T()));
}

/// convenience class to register a named object of the given class type with one constructor argument
template <class T, typename construct_arg_type>
object_registration_1<T,construct_arg_type>::object_registration_1(construct_arg_type arg)
{
	register_object(base_ptr(new T(arg)));
}

	}
}

#include <cgv/config/lib_end.h>