#pragma once

namespace cgv {
	namespace data {

/// different supported value types
enum DataType {
	DT_UNDEF  = -1,
	DT_BIT    = 0,
	DT_INT8, 
	DT_UINT8,
	DT_INT16, 
	DT_UINT16, 
	DT_INT32, 
	DT_UINT32, 
	DT_INT64, 
	DT_UINT64, 
	DT_FLT16,
	DT_FLT32,
	DT_FLT64,
};

enum DataComponent {
	DC_PX,
	DC_PY,
	DC_PZ,
	DC_PW,

	DC_NX,
	DC_NY,
	DC_NZ,

	DC_TU,
	DC_TV,
	DC_TS,
	DC_TW,
	
	DC_CR,
	DC_CG,
	DC_CB,
	DC_CA,

	DC_CH,
	DC_CL,
	DC_CS,

	DC_CX,
	DC_CY,
	DC_CZ,

	DC_CU,
	DC_CV,

	DC_CI,
	DC_CQ,

	DC_NR
};

struct data_component_info
{
	DataComponent component      : 8;
	unsigned int alignment       : 8;
	DataType      type           : 4;
	//unsigned int nr_step_sizes   : 4;
	//unsigned int first_step_size : 8;
};

struct data_format
{
	std::vector<data_component_info> components; // vector of maximally 256 components
	unsigned char component_idx[DC_NR];

	data_format(const char* text_descriptor);
};

class data_ptr
{
	data_format& format;
	unsigned char dim;
	unsigned char* data;
	unsigned short step_sizes[4];
public:
	data_ptr operator (unsigned int i) const
	{
		if (dim == 0) {
			std::cerr << "1d operator access to 0d data ptr" << std::endl;
			return data_ptr();
		}
		return data_ptr(format, dim-1, data+i*step_sizes[0],step_sizes+1);
	}
	data_ptr operator (unsigned int i, unsigned int j) const
	{
		if (dim < 2) {
			std::cerr << "2d operator access to " << dim << "d data ptr" << std::endl;
			return data_ptr();
		}
		return data_ptr(format, dim-2, data+i*step_sizes[0]+j*step_sizes[1],step_sizes+2);
	}
	data_ptr operator (unsigned int i, unsigned int j, unsigned int k) const
	{
		if (dim < 3) {
			std::cerr << "3d operator access to " << dim << "d data ptr" << std::endl;
			return data_ptr();
		}
		return data_ptr(format, dim-3, data+i*step_sizes[0]+j*step_sizes[1]+k*step_sizes[2],step_sizes+2);
	}

};

/** if you derive your class from this class, a ref_ptr will do reference counting in the inhereted 
    ref_count member. */
class ref_counted
{
private:
	/// keep a mutable reference count to allow ref counted points to const instances
	mutable int ref_count;
public:
	/// constructor initializes the count to 0
	ref_counted() : ref_count(0) {}
	/// read access to current count
	int get_ref_count() const       { return ref_count; }
	/// write access is also const to allow ref counted pointers to const instances
	void set_ref_count(int c) const { ref_count = c; }
};

	}
}