#include <cgv/gui/events.h>
#include <cgv/gui/fltk/gl/fltk_gl_trackball_interactor.h>

#define _USE_MATH_DEFINES
#include <cmath>
#include <cgv/render/gpu/gl/gl_tools.h>

using namespace cgv::gui::fltk;
using namespace cgv::render::gpu::gl;

namespace cgv {
	namespace gui {
		namespace fltk {
			namespace gl {


template <typename T>
cgv::math::linalg::fix_col_vec<T,3> rotate(
	const cgv::math::linalg::fix_col_vec<T,3>& v, 
	const cgv::math::linalg::fix_col_vec<T,3>& n, T a) 
{
	cgv::math::linalg::fix_col_vec<T,3> vn = dot(n,v)*n;
	return vn + cos(a)*(v-vn) + sin(a)*cross(n,v);
}


template <typename T>
trackball<T>::trackball()
{
	set_default_values();
}
template <typename T>
void trackball<T>::set_default_values()
{
	view_angle = 45;
	scale = 1;
	view_dir = v3d(0,0,1);
	focus = p3d(0,0,0);
	view_up_dir = v3d(0,1,0);
	z_near = (T)0.01;
	z_far  = (T)10000.0;
}

template <typename T>
typename trackball<T>::p3d trackball<T>::get_eye() const
{
	return focus-(scale/tan(.008726646260*view_angle))*view_dir;
}

template <typename T>
void trackball<T>::set_eye(const p3d& eye)
{
	view_dir = focus-eye;
	scale = tan(.008726646260*view_angle)*length(view_dir);
	view_dir = normalize(view_dir);
}

template <typename T>
void trackball<T>::put_coordinate_system(v3d& x, v3d& y, v3d& z) const
{
	z = normalize(-view_dir);
	x = normalize(cross(view_up_dir,z));
	y = cross(z,x);
}


void fltk_gl_trackball_interactor::focus_cb(fltk_key_double_control::event_type et, void* this_ptr, int i)
{
	((fltk_gl_trackball_interactor*) this_ptr)->update_cb(et,i);
}
void fltk_gl_trackball_interactor::ball_cb(fltk_key_double_control::event_type et, void* this_ptr, int i)
{
	((fltk_gl_trackball_interactor*) this_ptr)->update_cb(et,i);
}
void fltk_gl_trackball_interactor::roll_cb(fltk_key_double_control::event_type et, void* this_ptr, int i)
{
	((fltk_gl_trackball_interactor*) this_ptr)->update_cb(et,i);
}
void fltk_gl_trackball_interactor::scale_cb(fltk_key_double_control::event_type et, void* this_ptr, int i)
{
	((fltk_gl_trackball_interactor*) this_ptr)->update_cb(et,i);
}

void fltk_gl_trackball_interactor::update_cb(fltk_key_double_control::event_type et, int i)
{
	if (et == fltk_key_double_control::PREPARE)
		inc_action();
	else {
		update();
		if (et == fltk_key_double_control::FINISH)
			dec_action();
	}
}

void fltk_gl_trackball_interactor::inc_action()
{
	if (nr_actions == 0) {
		old_tb = *this;
		put_coordinate_system(x,y,z);
//		width = gl_get_width();
//		height = gl_get_height();
	}
	++nr_actions;
}
void fltk_gl_trackball_interactor::dec_action()
{
	if (--nr_actions == 0)
		init_delta_variables();
}

void fltk_gl_trackball_interactor::init_delta_variables()
{
	foc[0] = foc[1] = 0;
	ball_angle[0] = ball_angle[1] = 0;
	roll_angle = 0;
}

void fltk_gl_trackball_interactor::update()
{
	focus = old_tb.get_focus() - foc[0]*scale*x - foc[1]*scale*y;
	if (fabs(roll_angle) > 0.0001) {
		v3d n       = -z;
		double a    = roll_angle/60;
		view_up_dir = normalize(rotate(old_tb.get_view_up_dir(), n, a));
	}
	double a = sqrt(ball_angle[0]*ball_angle[0]+ball_angle[1]*ball_angle[1])/60;
	if (a > 0.0001) {
		v3d n = -ball_angle[0]*y + ball_angle[1]*x;
		n = normalize(n);
		view_dir    = normalize(rotate(old_tb.get_view_dir(), n, a));
		view_up_dir = normalize(rotate(old_tb.get_view_up_dir(), n, a));
	}
}

void fltk_gl_trackball_interactor::set_scene_extent(const b3d& b)
{
	scene_extent = b;
	scene_extent.add_point(b.get_center()+1.1*(b.maxp-b.get_center()));
	scene_extent.add_point(b.get_center()+1.1*(b.minp-b.get_center()));
	std::cout << "scene extent set to (" << b.minp[0] << ", " << b.minp[1] << ", " << b.minp[2] << "), ("
		<< b.maxp[0] << ", " << b.maxp[1] << ", " << b.maxp[2] << ")" << std::endl;
}

fltk_gl_trackball_interactor::fltk_gl_trackball_interactor() 
{
	aspect = 1;
	set_default_values();
	init_delta_variables();
	eye_distance = 0.03;
	stereo_frame = 0;
	nr_actions = 0;
	scene_extent.invalidate();

	add_object(new fltk_key_double_control(
		eye_distance, 0.05, 0.0, 0.2, 'b','b', true, ::fltk::SHIFT));

	add_object(new fltk_key_double_control(
		view_angle, 30.0, 3.0, 180.0, 'v','v', true, ::fltk::SHIFT));
	add_object(new fltk_key_double_control(
		z_near, 2.0, 'n','n', false, ::fltk::SHIFT));
	add_object(new fltk_key_double_control(
		z_far, 2.0, 'm','m', false, ::fltk::SHIFT));

	fltk_key_double_control* dc;
	add_object(dc = new fltk_key_double_control(
		foc[0], 2.0, ::fltk::LeftKey, ::fltk::RightKey));
	dc->set_callback(focus_cb, this, 0);
	add_object(dc = new fltk_key_double_control(
		foc[1], 2.0, ::fltk::DownKey, ::fltk::UpKey));
	dc->set_callback(focus_cb, this, 1);

	add_object(dc = new fltk_key_double_control(
		ball_angle[0], 100.0, ::fltk::LeftKey, ::fltk::RightKey, true, ::fltk::SHIFT, ::fltk::SHIFT));
	dc->set_callback(ball_cb, this, 0);
	add_object(dc = new fltk_key_double_control(
		ball_angle[1], 100.0, ::fltk::DownKey, ::fltk::UpKey, true, ::fltk::SHIFT, ::fltk::SHIFT));
	dc->set_callback(ball_cb, this, 1);

	add_object(dc = new fltk_key_double_control(
		roll_angle, 100.0, ::fltk::PageDownKey, ::fltk::PageUpKey, true, ::fltk::SHIFT, ::fltk::SHIFT));
	dc->set_callback(roll_cb, this);
	add_object(dc = new fltk_key_double_control(
		scale, 2.0, ::fltk::PageUpKey, ::fltk::PageDownKey, false));
	dc->set_callback(scale_cb, this);
}

const char* fltk_gl_trackball_interactor::get_name() const 
{
	return "fltk_gl_trackball_interactor"; 
}

void fltk_gl_trackball_interactor::show_stats(int& y) const 
{
	p3d e = get_eye();
	double va = view_angle;
	if (va <= 3)
		va = 0;
	gl_printf(20,y+=20,"nra=%d, View: angle=%.1f<v>, scale=%.1f, deye=%.3f, foc=%.2f,%.2f,%.2f, dir=%.2f,%.2f,%.2f, up=%.2f,%.2f,%.2f, z:[%.2f,%.2f]:[%.2g<n>, %.2g<m>]", 
		nr_actions,va,scale,eye_distance,focus[0],focus[1],focus[2],view_dir[0],view_dir[1],view_dir[2],
		view_up_dir[0],view_up_dir[1],view_up_dir[2],
		set_z_near, set_z_far, z_near, z_far);
}

void fltk_gl_trackball_interactor::scale_to_scene_extent()
{
	b3d b = scene_extent;
	std::cout << "scaling to scene extent with extent = (" << b.minp[0] << ", " << b.minp[1] << ", " << b.minp[2] << "), ("
		<< b.maxp[0] << ", " << b.maxp[1] << ", " << b.maxp[2] << ")" << std::endl;
	std::cout << "focus = " << focus[0] << ", " << focus[1] << ", " << focus[2] << std::endl;
	if (!scene_extent.is_valid())
		return;
	v3d x,y,z;
	b2d xy_extent;
	xy_extent.invalidate();
	put_coordinate_system(x,y,z);
	std::cout << "aspect = " << aspect << std::endl;
	for (unsigned int i=0; i<8; ++i) {
		v3d v  = scene_extent.get_corner(i)-focus;
		xy_extent.add_point(aspect*p2d(dot(v,x),dot(v,y)));
	}
	set_scale(
		__max(__max(xy_extent.maxp[0],xy_extent.maxp[1]),
		      __max(-xy_extent.minp[0],-xy_extent.minp[1]))
	);
	std::cout << "resulting scale: " << scale << std::endl;
}

void fltk_gl_trackball_interactor::init_frame() 
{
	width = gl_get_width();
	height = gl_get_height();
	aspect = (double)width/height;

	glLightModelf(GL_LIGHT_MODEL_LOCAL_VIEWER, 0.0f);
	// this command makes all successive matrix manipultations act on the projection transformation stack
	glMatrixMode(GL_PROJECTION);

/*	if (allow_parallel) {
		double aspect = (double)gl_get_width()/gl_get_height();
		double tany = tan(0.008726646262*tb.view_angle);
		double dF = tb.scale;
		double d  = 2*tb.scale;
		double sy = tb.scale;
		double sx = aspect*sy;

		double t4 = dF*tany-sy;
		double t5 = 1.0/(d*t4);
		double t6 = 1.0/t4;
		double A  = t5*(2*t4-d*tany);
		double B  = t6*tany;
		double D  = -t6*sy;
		double C  = t5*(2*d*dF*tany-2*dF*t4-d*sy);
		double P[16] = {
			-1/sx,0,   0,0,
			0,   1/sy,0,0,
			0,   0,   A,B,
			0,   0,   C,D
		};
		glLoadMatrixd(P);
	}
	else {
	*/
		// initialize the projective transformation to the identity
		glLoadIdentity();
		// define a viewing pyramid with an opening angle of 45 degrees in the y-direction and 
		// an aspect ration computed from the fltk-window sizes w() and h(). Set the near clipping
		// plane to 0.5 and the far clipping plane to 10
		set_z_near = z_near;
		set_z_far  = z_far;
		if (scene_extent.is_valid()) {
			double z_min = dot(scene_extent.get_corner(0),view_dir);
			double z_max = z_min;
			for (unsigned int i=1; i<8; ++i) {
				double new_z= dot(scene_extent.get_corner(i),view_dir);
				if (new_z< z_min)
					z_min = new_z;
				if (new_z > z_max)
					z_max = new_z;
			}
			double z_eye = dot(get_eye(),view_dir);
			z_min -= z_eye;
			z_max -= z_eye;
			if (z_min > z_near)
				set_z_near = z_min;
			if (z_max > z_near && z_max < z_far)
				set_z_far = z_max;
		}
		double aspect = (double)gl_get_width()/gl_get_height();
		if (view_angle <= 3)
			glOrtho(-aspect*scale, aspect*scale, -scale, scale, set_z_near, set_z_far);
		else {
//			gluPerspective(view_angle, aspect, set_z_near, set_z_far);
			double top    = set_z_near*tan(0.008726646262*view_angle);
			double bottom = -top;
			double delta  = stereo_frame*top*aspect*eye_distance;
			double left   = bottom*aspect-delta;
			double right  = top*aspect-delta;
			glFrustum(left,right,bottom,top,set_z_near,set_z_far);
		}
//	}
	// switch back to the modelview transformation stack
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
/*
	if (allow_parallel) {
		p3d foc = tb.focus+tb.eye_dir;
		gluLookAt(tb.focus[0],tb.focus[1],tb.focus[2],
					 foc[0], foc[1], foc[2],
					 tb.up[0],tb.up[1],tb.up[2]);
	}
	else {
	*/
		// set up the observer at location (0,-3,4) and look at the focus point (0,1,-1) with
		// the upward direction of the image along the z-axis
		v3d x, y, z;
		put_coordinate_system(x,y,z);
		double s = scale*aspect*stereo_frame*eye_distance;
		p3d eye = get_eye() + s*x;
		p3d f   = focus + s*x;
		gluLookAt(eye[0],eye[1],eye[2],f[0],f[1],f[2],view_up_dir[0],view_up_dir[1],view_up_dir[2]);
//	}
}

void fltk_gl_trackball_interactor::show_help(int& y) const 
{
	gl_printf(20,y+=20,"VIEW: Focus  ... move: arrows/left button;             ... set:     Ctrl-left_button");
	gl_printf(20,y+=18,"      Zoom   ... change: PgUp,PgDn / wheel;            ... init:               Space");
	gl_printf(20,y+=18,"      Rotate ... free: Shift+arrows/Shift-left_button; ... constrained: right_button");
	gl_printf(20,y+=18,"      Roll   ... Shift-PgUp,PgDn / Shift-right_button");
}

fltk_gl_trackball_interactor::Result fltk_gl_trackball_interactor::handle_event(int event) 
{
	if (event == SELECT_STEREO_FRAME) {
		stereo_frame = event_data<int>();
		return HANDLED_AND_REDRAW;
	}
	if (event == SCALE_EVENT) {
		set_scale(event_data<double>());
		return HANDLED_AND_REDRAW;
	}
	if (event == SET_FOCUS_POINT) {
		event_point& p = event_data<event_point>();
		set_focus(p3d(p[0],p[1],p[2]));
		return HANDLED_AND_REDRAW;
	}
	if (event == SET_EYE_POINT) {
		event_point& p = event_data<event_point>();
		set_eye(p3d(p[0],p[1],p[2]));
		return HANDLED_AND_REDRAW;
	}
	if (event == SET_VIEW_DIR) {
		event_vec& v = event_data<event_vec>();
		set_view_dir(v3d(v[0],v[1],v[2]));
		return HANDLED_AND_REDRAW;
	}
	if (event == SET_UP_DIR) {
		event_vec& v = event_data<event_vec>();
		set_view_up_dir(v3d(v[0],v[1],v[2]));
		return HANDLED_AND_REDRAW;
	}
	if (event == SET_FIELD_OF_VIEW) {
		view_angle = event_data<double>();
		return HANDLED_AND_REDRAW;
	}
	if (event == SET_Z_NEAR) {
		z_near = event_data<double>();
		return HANDLED_AND_REDRAW;
	}
	if (event == SET_Z_FAR) {
		z_far = event_data<double>();
		return HANDLED_AND_REDRAW;
	}
	if (event == SET_SCALE) {
		set_scale(event_data<double>());
		return HANDLED_AND_REDRAW;
	}



	if (event == GET_FOCUS_POINT) {
		event_point& p = event_data<event_point>();
		p[0] = focus[0];
		p[1] = focus[1];
		p[2] = focus[2];
		return HANDLED;
	}
	if (event == GET_EYE_POINT) {
		event_point& p = event_data<event_point>();
		p[0] = get_eye()[0];
		p[1] = get_eye()[1];
		p[2] = get_eye()[2];
		return HANDLED;
	}
	if (event == GET_VIEW_DIR) {
		event_vec& v = event_data<event_vec>();
		v[0] = view_dir[0];
		v[1] = view_dir[1];
		v[2] = view_dir[2];
		return HANDLED;
	}
	if (event == GET_UP_DIR) {
		event_vec& v = event_data<event_vec>();
		v[0] = view_up_dir[0];
		v[1] = view_up_dir[1];
		v[2] = view_up_dir[2];
		return HANDLED;
	}
	if (event == GET_FIELD_OF_VIEW) {
		event_data<double>() = view_angle;
		return HANDLED;
	}
	if (event == GET_Z_NEAR) {
		event_data<double>() = z_near;
		return HANDLED;
	}
	if (event == GET_Z_FAR) {
		event_data<double>() = z_far;
		return HANDLED;
	}
	if (event == GET_SCALE) {
		event_data<double>() = scale;
		return HANDLED_AND_REDRAW;
	}



	if (event == SET_SCENE_EXTENT) {
		event_box& b = event_data<event_box>();
		set_scene_extent(b3d(p3d(b.minp[0],b.minp[1],b.minp[2]),
									p3d(b.maxp[0],b.maxp[1],b.maxp[2])));
		return HANDLED_AND_REDRAW;
	}
	if (event == ADD_SCENE_EXTENT) {
		event_box& b = event_data<event_box>();
		b3d B(p3d(b.minp[0],b.minp[1],b.minp[2]),
				p3d(b.maxp[0],b.maxp[1],b.maxp[2]));
		scene_extent.add_point(B.get_center()+1.1*(B.minp-B.get_center()));
		scene_extent.add_point(B.get_center()+1.1*(B.maxp-B.get_center()));
		return HANDLED_AND_REDRAW;
	}
	if (event == SCALE_TO_SCENE_EXTENT) {
		scale_to_scene_extent();
		return HANDLED_AND_REDRAW;
	}


	if (event == cgv::gui::DOF6_SENSOR) {
		cgv::gui::event_vec& t = cgv::gui::event_data<cgv::gui::event_vec>();
		inc_action();
		foc[0]+= 0.2*t[0];
		foc[1]-= 0.2*t[2];
		std::cout << "transl: " << t[0] << ", " << t[1] << std::endl;
		scale *= exp(-0.1*t[1]);
		update();
		dec_action();
		return HANDLED_AND_REDRAW;
	}

	if (::fltk::event_state() == 0 && event == ::fltk::KEY && ' ' == ::fltk::event_key()) {
		if (scene_extent.is_valid()) {
			set_focus(scene_extent.get_center());
			scale_to_scene_extent();
		}
		else {
			set_default_values();
			init_delta_variables();
		}
		return HANDLED_AND_REDRAW;
	}
	if (event == ::fltk::MOUSEWHEEL) {
		double s = 0.1;
		if (::fltk::event_state() == ::fltk::SHIFT)
			s = 0.02;
		else if (::fltk::event_state() == ::fltk::CTRL)
			s = 0.4;
		scale *= exp(s*::fltk::event_dy());
		return HANDLED_AND_REDRAW;
	}
	if (event == ::fltk::PUSH && 0 == (::fltk::event_state()&(::fltk::ALT+::fltk::CTRL))) {
		switch (::fltk::event_button()) {
		case ::fltk::LeftButton :
		case ::fltk::RightButton :
			inc_action();
			last_x = ::fltk::event_x();
			last_y = ::fltk::event_y();
			return HANDLED;
		}
	}
	if (event == ::fltk::PUSH && ::fltk::event_state() == (::fltk::CTRL|::fltk::BUTTON1)) {
		event_point& p = event_data<event_point>();
//		std::cout << "set focus to " << p[0] << "," << p[1] << "," << p[2] << std::endl;
		set_focus(p3d(p[0],p[1],p[2]));
		return HANDLED_AND_REDRAW;
	}
	if (event == ::fltk::RELEASE && ::fltk::event_state() == (::fltk::CTRL|::fltk::BUTTON1))
		return HANDLED;

	if (event == ::fltk::RELEASE && 0 == (::fltk::event_state()&(::fltk::ALT+::fltk::CTRL))) {
		switch (::fltk::event_button()) {
		case ::fltk::LeftButton :
		case ::fltk::RightButton :
			dec_action();
			return HANDLED_AND_REDRAW;
		}
	}
	if (event == ::fltk::DRAG && 0 == (::fltk::event_state()&(::fltk::ALT+::fltk::CTRL))) {
		int dx = ::fltk::event_x() - last_x;
		int dy = ::fltk::event_y() - last_y;
		last_x = ::fltk::event_x();
		last_y = ::fltk::event_y();
		switch (::fltk::event_state()) {
		case ::fltk::BUTTON3 :
			foc[0] += 2.0*dx/width;
			foc[1] -= 2.0*dy/height;
			update();
			return HANDLED_AND_REDRAW;
		case ::fltk::BUTTON1 :
			ball_angle[0] += 265.0*dx/width;
			ball_angle[1] -= 265.0*dy/height;
			update();
			return HANDLED_AND_REDRAW;
		case ::fltk::BUTTON3|::fltk::SHIFT :
			{
				int ax = dx<0?-dx:dx;
				int ay = dy<0?-dy:dy;
				if (ax+ay < 8) {
					last_x -= dx;
					last_y -= dy;
					return HANDLED;
				}
				if (ay != 0 && (float)ax/ay < 1.8f && ax != 0 && (float)ay/ax < 1.8f)
					return HANDLED;
				if (ax>ay)
					ball_angle[0] += 265.0*dx/width;
				else
					ball_angle[1] -= 265.0*dy/height;
				update();
				return HANDLED_AND_REDRAW;
			}
		case ::fltk::BUTTON1|::fltk::SHIFT :
			{
				int rx = last_x - width/2;
				int ry = last_y - height/2;
				double ds = sqrt(((double)dx*(double)dx+(double)dy*(double)dy)/((double)rx*(double)rx+(double)ry*(double)ry));
				if (rx*dy > ry*dx)
					ds = -ds;
				roll_angle += 56.3*ds;
			}
			update();
			return HANDLED_AND_REDRAW;
		}
	}
	return named_object_vector::handle_event(event);
}

			}
		}
	}
}
