#pragma once

#include <cgv/math/linalg/linalg.h>
#include <cgv/scene/geometry/point.h>
#include <cgv/scene/geometry/axis_aligned_box.h>
#include <cgv/gui/fltk/named_object_vector.h>
#include <cgv/gui/fltk/fltk_key_double_control.h>

#include <cgv/gui/fltk/lib_begin.h>

namespace cgv {
	namespace gui {
		namespace fltk {
			namespace gl {


template <typename T>
class CGV_API trackball
{
public:
	typedef cgv::scene::geometry::point<T,3> p3d;
	typedef cgv::math::linalg::fix_col_vec<T,3> v3d;
protected:
	p3d focus;
	v3d view_dir;
	v3d view_up_dir;
	T view_angle;
	T scale;
	T z_near;
	T z_far;
public:
	trackball();
	void set_default_values();
	p3d get_eye() const;
	void set_eye(const p3d& e);
	T get_view_angle() const { return view_angle; }
	void set_view_angle(T a) { view_angle = v; }
	T get_z_near() const { return z_near; }
	void set_z_near(T z) { z_near = z; }
	T get_z_far() const { return z_far; }
	void set_z_far(T z) { z_far = z; }
	const p3d& get_focus() const { return focus; }
	void set_focus(const p3d& f) { focus = f; }
	void set_scale(T s) { scale = s; }
	T get_scale() const { return scale; }
	const v3d& get_view_dir() const { return view_dir; }
	void set_view_dir(const v3d& v) { view_dir = normalize(v); }
	const v3d& get_view_up_dir() const { return view_up_dir; }
	void set_view_up_dir(const v3d& v) { view_up_dir = normalize(v); }
	void put_coordinate_system(v3d& x, v3d& y, v3d& z) const;
};


class CGV_API fltk_gl_trackball_interactor : 
	public named_object_vector, 
	public trackball<double>
{
public:
	typedef trackball<double> tb_type;
	typedef cgv::scene::geometry::axis_aligned_box<double,3> b3d;
	typedef cgv::scene::geometry::axis_aligned_box<float,3> b3f;
	typedef cgv::scene::geometry::axis_aligned_box<double,2> b2d;
	typedef cgv::scene::geometry::point<double,2> p2d;
	typedef cgv::scene::geometry::point<float,3> p3f;
	typedef cgv::math::linalg::fix_col_vec<float,3> v3f;
	
	static void focus_cb(fltk_key_double_control::event_type et, void* this_ptr, int i);
	static void ball_cb(fltk_key_double_control::event_type et, void* this_ptr, int i);
	static void roll_cb(fltk_key_double_control::event_type et, void* this_ptr, int i);
	static void scale_cb(fltk_key_double_control::event_type et, void* this_ptr, int i);
private:
	double aspect;
protected:
	void update_cb(fltk_key_double_control::event_type et, int i);
	double eye_distance;
	int stereo_frame;
	/// count current number of actions
	int nr_actions;
	void inc_action();
	void dec_action();

	/// reference trackball and coordinate system
	tb_type old_tb;
	v3d x, y, z;

	/// incremental update variables
	double foc[2];
	double roll_angle;
	double ball_angle[2];

	void init_delta_variables();
	void update();

	/// mouse interaction variables
	int last_x, last_y, width, height;

	///
	b3d    scene_extent;
	double set_z_near, set_z_far;	
public:
	fltk_gl_trackball_interactor();
	const char* get_name() const;
	void set_scene_extent(const b3d& b);
	const b3d& get_scene_extent() const { return scene_extent; }
	void scale_to_scene_extent();
	void show_stats(int& y) const;
	void init_frame();
	void show_help(int& y) const;
	Result handle_event(int event);
};
			}
		}
	}
}

#include <cgv/config/lib_end.h>