#include <iostream>
#include <cgv/gui/fltk/gl_view.h>
#include <cgv/gui/fltk/fltk_font_server.h>
#include <cgv/render/drawable.h>
#include <fltk/events.h>
#include <fltk/run.h>
#include <fltk/draw.h>
#include <fltk/Font.h>
#include <fltk/gl.h>
#include <fltk/visual.h>
#include <cgv/media/image/bmp.h>
#include <cgv/media/text/ostream_printf.h>
#include <cgv/media/text/scan.h>
#include <cgv/utils/clipboard.h>
#include <cgv/gui/key_event.h>
#include <cgv/gui/mouse_event.h>
#ifdef WIN32
#include <windows.h>
#endif
#include <gl/gl.h>
#include <gl/glu.h>

using namespace cgv::base;
using namespace cgv::render;
using namespace cgv::render::gl;

namespace cgv {
	namespace gui {
		namespace fltk {

const int nr_backgrounds = 5;
GLclampf background_colors[] = {
	0,0,0,0,
	1.0f, 0.4f, 0.5f,0,
	0.4f, 1, 0.7f,0,
	0.5f, 0.5f, 1,0,
	1,1,1,0,
};


/// construct application
gl_view::gl_view(int x, int y, int w, int h, const char* name) : ::fltk::GlWindow(x,y,w,h), group(name)
{
	current_background = 0;
	bg_r = bg_g = bg_b = bg_a = 0;	
	x_offset = 10;
	y_offset = 20;
	tab_size = 10;
	text_x = x_offset;
	text_y = y_offset;
	nr_identations = 0;
	at_line_begin = true;

	last_context = (void*)-1;
	last_width = -1;
	last_height = -1;

	in_draw_method = false;

	current_font_size = 14;

	info_font_size = 14;
	// check if a new context has been created or if the size of the viewport has been changed
	media::font::font_ptr info_font = media::font::find_font("Courier New");
	if (info_font.empty()) {
		info_font = media::font::find_font("Courier");
		if (info_font.empty()) {
			info_font = media::font::find_font("system");
		}
	}
	if (!info_font)
		std::cerr << "could not find a font" << std::endl;
	else {
		info_font_face = info_font->get_font_face(media::font::FFA_REGULAR);
	}

	connect(out_stream.write, this, &gl_view::process_text);

	show_help = false;
	show_stats = true;

	do_screen_shot = false;

	if ( (mode() & ::fltk::ALPHA_BUFFER) == 0 ) {
		std::cout << "turn on alpha buffer" << std::endl;
		mode(mode()|::fltk::ALPHA_BUFFER);
	}
}

void gl_view::destroy()
{
	remove_all_children();
	::fltk::GlWindow::destroy();
}

/// helper method to integrate a new child
void gl_view::configure_new_child(base_ptr child)
{
	traverser(make_action<render::context*>(this, &drawable::set_context)).traverse(child);
	traverser(make_action<render::context&>(*this, &drawable::init),"nc").traverse(child);
	redraw();
}

/// helper method to remove a child
void gl_view::clear_child(base_ptr child)
{
	traverser(make_action<render::context&>(*this, &drawable::clear),"nc").traverse(child);
}

/// append child and return index of appended child
unsigned int gl_view::append_child(base_ptr child)
{
	unsigned int i = group::append_child(child);
	configure_new_child(child);
	set_focused_child(i);
	return i;
}
/// remove all elements of the vector that point to child, return the number of removed children
unsigned int gl_view::remove_child(base_ptr child)
{
	clear_child(child);
	return group::remove_child(child);
}
/// remove all children
void gl_view::remove_all_children()
{
	for (unsigned int i=0; i<get_nr_children(); ++i)
		clear_child(get_child(i));
	group::remove_all_children();
}
/// insert a child at the given position
void gl_view::insert_child(unsigned int i, base_ptr child)
{
	group::insert_child(i,child);
	configure_new_child(child);
	set_focused_child(i);
}


/// complete drawing method that configures opengl whenever the context has changed, initializes the current frame and draws the scene
void gl_view::draw()
{
	in_draw_method = true;
	if (!valid())
		configure_gl();
	init_frame();

	traverser(make_action<render::context&>(*this, &drawable::draw, &drawable::finish_draw)).traverse(group_ptr(this));
	traverser(make_action<render::context&>(*this, &drawable::finish_frame)).traverse(group_ptr(this));
	draw_textual_info();
	if (do_screen_shot) {
		glFlush();
		unsigned char *buffer = new unsigned char[w()*h()*3];
		glPixelStorei(GL_PACK_ALIGNMENT, 1);
		glReadPixels(0,0,w(),h(),GL_RGB,GL_UNSIGNED_BYTE,buffer);
		cgv::utils::copy_rgb_image_to_clipboard(w(),h(),buffer);
		cgv::media::image::write_rgb_bmp("screen_shot.bmp",w(),h(),buffer);
		delete [] buffer;
		do_screen_shot = false;
	}

	in_draw_method = false;
}

/// put the event focus on the given child 
bool gl_view::set_focus(base_ptr child)
{
	for (unsigned int i = 0; i < get_nr_children(); ++i)
		if (child == get_child(i)) {
			set_focused_child(i);
			if (show_stats)
				redraw();
			return true;
		}
	return false;
}

/// handle events emitted by event sources
bool gl_view::handle(event& e)
{
	if (e.get_kind() == EID_KEY) {
		key_event& ke = static_cast<key_event&>(e);
		if (ke.get_action() == KA_PRESS) {
			switch (ke.get_key()) {
			case KEY_F4 : 
				if (ke.get_modifiers() == EM_ALT) {
					destroy();
					exit(0);
				}
				break;
			case KEY_F1 : 
				show_help = !show_help;
				redraw();
				break;
			case KEY_Delete :
				if (get_focused_child() != -1) {
					remove_child(get_child(get_focused_child()));
					if (get_focused_child() >= (int)get_nr_children())
						traverse_policy::set_focused_child(get_focused_child()-1);
					redraw();
				}
				break;
			case KEY_Tab :
				if (ke.get_modifiers() == EM_SHIFT) {
					traverse_policy::set_focused_child(get_focused_child()-1);
					if (get_focused_child() == -2)
						traverse_policy::set_focused_child((int)get_nr_children()-1);
				}
				else if (ke.get_modifiers() == 0) {
					traverse_policy::set_focused_child(get_focused_child()+1);
					if (get_focused_child() >= (int)get_nr_children())
						traverse_policy::set_focused_child(-1);
				}
				if (show_stats)
					redraw();
				break;
			case KEY_F8 : 
				show_stats = !show_stats; 
				redraw();
				break;
			case KEY_F10 : 
				do_screen_shot = true;
				redraw();
				break;
			case KEY_F12 :
				if (++current_background >= nr_backgrounds)
					current_background = 0;
				set_bg_color(background_colors[4*current_background],background_colors[4*current_background+1],background_colors[4*current_background+2],background_colors[4*current_background+3]);
				break;
			default: 
				return false;
			}
			return true;
		}
	}
	return false;
}
/// overload to stream help information to the given output stream
void gl_view::stream_help(std::ostream& os)
{
	os << "MAIN: Show   ... help: F1;                  ... status: F8;\n";
	os << "      Object ... change focus: [Shift-]Tab; ... remove focused: Delete\n";
	os << "      Screen ... copy to file: F10;         ... background: F12\n";
	os << "Quit: Alt-F4\n" << std::endl;
}

void gl_view::stream_stats(std::ostream& os)
{
	std::string name("no focus");
	if (get_focused_child() != -1) {
		base_ptr c = get_child(get_focused_child());
		named* n = dynamic_cast<named*>(c.ref());
		if (n) {
			name  = n->get_name();
			name += ':';
		}
		else
			name = "";
		name += c->get_type_name();
	}
	media::text::oprintf(os, "focus = %s<Tab>\n", name.c_str());
}

struct format_callback_handler : public traverse_callback_handler
{
	std::ostream& os;
	format_callback_handler(std::ostream& _os) : os(_os) 
	{
	}
	/// called before the children of a group node g are processed, return whether these should be skipped. If children are skipped, the on_leave_children callback is still called.
	bool on_enter_children(group* g)
	{
		os << "\a";
		return false;
	}
	/// called when the children of a group node g have been left, return whether to terminate traversal
	bool on_leave_children(group* g)
	{
		os << "\b";
		return false;
	}

};

void gl_view::draw_textual_info()
{
	if (show_help || show_stats) {
		glDisable(GL_LIGHTING);
		glDisable(GL_DEPTH_TEST);
		   if (bg_r+bg_g+bg_b < 1.5f)
				glColor4f(1,1,1,1);
			else
				glColor4f(0,0,0,1);

			push_pixel_coords();
			enable_font_face(info_font_face, info_font_size);
			format_callback_handler fch(output_stream());
			set_cursor(20,20);
			if (show_stats) {
				traverser(make_action<std::ostream&>(output_stream(), &base::stream_stats, false),"nc").traverse(base_ptr(this),&fch);
				output_stream() << std::endl;
			}
		    if (bg_r+bg_g+bg_b < 1.5f)
				glColor4f(1,1,0,1);
			else
				glColor4f(0.4f,0.3f,0,1);
			if (show_help) {
				// collect help from myself and all children
				traverser(make_action<std::ostream&>(output_stream(), &event_handler::stream_help, false),"nc").traverse(base_ptr(this),&fch);
				output_stream().flush();
			}
			pop_pixel_coords();
		// turn lighting back on for the next frame
		glEnable(GL_DEPTH_TEST);
		glEnable(GL_LIGHTING);
	}
}

/// return the width of the window
unsigned int gl_view::get_width() const
{
	return w();
}
/// return the height of the window
unsigned int gl_view::get_height() const
{
	return h();
}

/// set a user defined background color
void gl_view::set_bg_color(float r, float g, float b, float a)
{
	bg_r = r;
	bg_g = g;
	bg_b = b;
	bg_a = a;
	redraw();
}

/// return the current render pass
RenderPass gl_view::get_render_pass() const
{
	if (render_pass_stack.empty())
		return RP_NONE;
	return render_pass_stack.top().pass;
}
/// return the current render pass flags
RenderPassFlags gl_view::get_render_pass_flags() const
{
	if (render_pass_stack.empty())
		return RPF_NONE;
	return render_pass_stack.top().flags;
}

/// perform the given render task
void gl_view::render_pass(RenderPass _render_pass, RenderPassFlags _render_pass_flags, bool call_init_frame, bool call_finish_frame)
{
	render_info ri;
	ri.pass  = _render_pass;
	ri.flags = _render_pass_flags;

	render_pass_stack.push(ri);
		if (call_init_frame)
			init_frame();

		traverser(make_action<render::context&>(*this, &drawable::draw, &drawable::finish_draw)).traverse(group_ptr(this));

		if (call_finish_frame)
			traverser(make_action<render::context&>(*this, &drawable::finish_frame)).traverse(group_ptr(this));
	render_pass_stack.pop();
}
/// return whether the context is currently in process of rendering
bool gl_view::in_render_process() const
{
	return in_draw_method;
}
/// return whether the context is current
bool gl_view::is_current() const
{
	return false;
}
/// make the current context current
void gl_view::make_current() const
{
	const_cast<::fltk::GlWindow*>(static_cast<const ::fltk::GlWindow*>(this))->make_current();
}

/// the context will be redrawn when the system is idle again
void gl_view::post_redraw()
{
	redraw();
}


void gl_view::force_redraw()
{
	if (in_render_process()) {
		std::cerr << "force_redraw not allowed in render process" << std::endl;
		post_redraw();
		return;
	}
	flush();
	set_damage(0);
}
/// select a font given by a font handle
void gl_view::enable_font_face(media::font::font_face_ptr font_face, float font_size)
{
	if (font_face.ref() != current_font_face.ref() || font_size != current_font_size) {
		const fltk_font_face* fff = dynamic_cast<const fltk_font_face*>(font_face.ref());
		if (!fff) {
			std::cerr << "could not use font face together with fltk" << std::endl;
			return;
		}
		::fltk::glsetfont(fff->get_fltk_font(),font_size);
		current_font_face = font_face;
		current_font_size = font_size;
	}
}
/// return the size in pixels of the currently enabled font face
float gl_view::get_current_font_size() const
{
	return current_font_size;
}
/// return the currently enabled font face
media::font::font_face_ptr gl_view::get_current_font_face() const
{
	return current_font_face;
}

/// overload to set text parameters
void gl_view::set_cursor(int x, int y)
{
	/// current x and y position of text
	text_x = x;
	text_y = y;
	x_offset = x;
	y_offset = y;
	nr_identations = 0;
	at_line_begin = true;

	render::gl::context::set_cursor(x,y);
}

/// returns an output stream whose output is printed at the current cursor location
std::ostream& gl_view::output_stream()
{
	return out_stream;
}

/// callback method for processing of text from the output stream
void gl_view::process_text(const std::string& text)
{
	// use fltk to draw the text
	render::gl::context::set_cursor(text_x,text_y);
	unsigned int i, j = 0;
	for (i = 0; i<text.size(); ++i) {
		int n = i-j;
		switch (text[i]) {
		case '\a' :
			if (n) {
				::fltk::gldrawtext(&text[j], n);
				get_cursor(text_x, text_y);
			}
			++nr_identations;
			if (at_line_begin) {
				text_x += tab_size;
				render::gl::context::set_cursor(text_x,text_y);
			}
			j = i+1;
			break;
		case '\b' :
			if (n) {
				::fltk::gldrawtext(&text[j], n);
				get_cursor(text_x, text_y);
			}
			if (nr_identations > 0) {
				--nr_identations;
				if (at_line_begin) {
					text_x -= tab_size;
					render::gl::context::set_cursor(text_x,text_y);
				}
			}
			j = i+1;
			break;
		case '\t' :
			if (n) {
				::fltk::gldrawtext(&text[j], n);
				get_cursor(text_x, text_y);
			}
			text_x = (((text_x-x_offset)/tab_size)+1)*tab_size;
			render::gl::context::set_cursor(text_x,text_y);
			at_line_begin = false;
			j = i+1;
			break;
		case '\n' :
			if (n) {
				::fltk::gldrawtext(&text[j], n);
				get_cursor(text_x, text_y);
			}
			text_x = x_offset+nr_identations*tab_size;
			text_y = text_y + (int)(1.2f*current_font_size);
			render::gl::context::set_cursor(text_x,text_y);
			at_line_begin = true;
			j = i+1;
			break;
		default:
			at_line_begin = false;
		}
	}	
	if (i-j > 1) {
		::fltk::gldrawtext(&text[j], i-j);
		get_cursor(text_x, text_y);
	}
}

/// define lighting mode, viewing pyramid and the rendering mode
void gl_view::configure_gl()
{
	if (GlWindow::context() != last_context) {
		last_context = GlWindow::context();
		enable_font_face(info_font_face,info_font_size);
		// use the eye location to compute the specular lighting
		glLightModelf(GL_LIGHT_MODEL_LOCAL_VIEWER, 0);
		glEnable(GL_NORMALIZE);
		glViewport(0,0,w(),h());
		traverser(make_action<render::context*>(this, &drawable::set_context)).traverse(group_ptr(this));
		traverser(make_action<render::context&>(*this, &drawable::init)).traverse(group_ptr(this));
	}
	else if (last_width != w() || last_height != h()) {
		last_width  = w();
		last_height = h();
		glViewport(0,0,w(),h());
		traverser(make_action_2<unsigned int,unsigned int>(last_width, last_height, &drawable::resize)).traverse(group_ptr(this));
	}
}

void gl_view::init_frame()
{
	// this defines the background color to which the frame buffer is set by glClear
	glClearColor(bg_r,bg_g,bg_b,bg_a);

	if ((get_render_pass_flags()&(RPF_LIGHTS_SET|RPF_NO_LIGHTING)) == 0) {
		// define diffuse and specular components of 4 light sources to gray and white
		// if the diffuse components where also white the 4 light contributions would
		// generate colors larger than one
		glLightfv(GL_LIGHT0, GL_DIFFUSE, white);
		glLightfv(GL_LIGHT1, GL_DIFFUSE, white);
		glLightfv(GL_LIGHT2, GL_DIFFUSE, white);
		glLightfv(GL_LIGHT3, GL_DIFFUSE, white);
		glLightfv(GL_LIGHT0, GL_SPECULAR, white);
		glLightfv(GL_LIGHT1, GL_SPECULAR, white);
		glLightfv(GL_LIGHT2, GL_SPECULAR, white);
		glLightfv(GL_LIGHT3, GL_SPECULAR, white);
		// enable lighting calculations
		glEnable(GL_LIGHTING);
		// this makes opengl normalize all surface normals before lighting calculations,
		// which is essential when using scaling to deform tesselated primities
		glEnable(GL_NORMALIZE);
		// turn the 4 lights on
		glEnable(GL_LIGHT0);
//		glEnable(GL_LIGHT1);
//		glEnable(GL_LIGHT2);
//		glEnable(GL_LIGHT3);

		// set the surface material colors and the specular exponent,
		// which is between 0 and 128 and generates sharpest highlights for 128 and no
		// distinguished highlight for 0
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, black);
		glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, white);
		glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, black);
		glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 80);
		// this mode allows to define the ambient and diffuse color of the surface material
		// via the glColor commands
		glEnable(GL_COLOR_MATERIAL);
	}
	// enable the depth test in order to show only the front most elements of the scene in 
	// each pixel
	glEnable(GL_DEPTH_TEST);
	// cull away the faces whose orientation points away from the observer
	glCullFace(GL_BACK);
	glEnable(GL_CULL_FACE);

	if ((get_render_pass_flags()&RPF_PROJECTION_SET) == 0) {
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective(45,(double)w()/h(),0.001,1000.0);
	}
	glMatrixMode(GL_MODELVIEW);
	if ((get_render_pass_flags()&RPF_MODELVIEW_SET) == 0) {
		glLoadIdentity();
		gluLookAt(0,0,10, 0,0,0, 0,1,0);
	}
	if ((get_render_pass_flags()&(RPF_LIGHTS_SET|RPF_NO_LIGHTING)) == 0) {
		float lps[] = { 0,1,1,0, 1,1,0,0, 0,0,1,0, 0,1,0,0 };
		glLightfv(GL_LIGHT0, GL_POSITION, lps);
		glLightfv(GL_LIGHT1, GL_POSITION, lps+4);
		glLightfv(GL_LIGHT2, GL_POSITION, lps+8);
		glLightfv(GL_LIGHT3, GL_POSITION, lps+12);
	}
	traverser(make_action<render::context&>(*this, &drawable::init_frame)).traverse(group_ptr(this));
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
}

unsigned char make_modifiers()
{
	unsigned char modifiers = 0;
	if (::fltk::event_state() & ::fltk::SHIFT) 
		modifiers += EM_SHIFT;
	if (::fltk::event_state() & ::fltk::ALT) 
		modifiers += EM_ALT;
	if (::fltk::event_state() & ::fltk::CTRL) 
		modifiers += EM_CTRL;
	if (::fltk::event_state() & ::fltk::META) 
		modifiers += EM_META;
	return modifiers;
}

unsigned char make_toggle_keys()
{
	unsigned char t = 0;
	if (::fltk::event_state() & ::fltk::CAPSLOCK) 
		t += ETK_CAPS_LOCK;
	if (::fltk::event_state() & ::fltk::NUMLOCK) 
		t += ETK_NUM_LOCK;
	if (::fltk::event_state() & ::fltk::SCROLLLOCK) 
		t += ETK_SCROLL_LOCK;
	return t;
}

unsigned short* get_key_translation_table()
{
	static unsigned short key_translation_table[256] =
	{
		0,0,0,0,0,0,0,0,KEY_Back_Space,KEY_Tab,0,0,0,KEY_Enter,0,0,
		0,0,0,KEY_Pause,KEY_Scroll_Lock,0,0,0,0,0,0,KEY_Escape,0,0,0,0,
		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
		KEY_Home, KEY_Left, KEY_Up, KEY_Right, KEY_Down, KEY_Page_Up, KEY_Page_Down, KEY_End, 0,0,0,0,0,0,0,0,
		0, KEY_Print, 0, KEY_Insert, 0,0,0,0,0,0,0,0,0,0,0,0,
		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,KEY_Num_Lock,
		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
		0,0,0,0,0,0,0,0,0,0,0,0,0,0,KEY_F1,KEY_F2,KEY_F3,
		KEY_F4,KEY_F5,KEY_F6,KEY_F7,KEY_F8,KEY_F9,KEY_F10,KEY_F11,KEY_F12,0,0,0,0,0,0,0,
		0,0,0,0,0,0,0,0,0,0,0,0,0,0,'A','B',
		KEY_Left_Shift,KEY_Right_Shift,KEY_Left_Ctrl,KEY_Right_Ctrl,KEY_Caps_Lock,0,KEY_Left_Meta,KEY_Right_Meta,KEY_Left_Alt,KEY_Right_Alt,'D','E','F','G','H',
		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,KEY_Delete
	};
	static bool initialized = false;
	if (!initialized) {
		initialized = true;
		key_translation_table[::fltk::KeypadEnter-0xff00] = KEY_Num_Enter;
		key_translation_table[::fltk::MultiplyKey-0xff00] = KEY_Num_Mul;
		key_translation_table[::fltk::AddKey-0xff00]      = KEY_Num_Add;
		key_translation_table[::fltk::SubtractKey-0xff00] = KEY_Num_Sub;
		key_translation_table[::fltk::DecimalKey-0xff00]  = KEY_Num_Dot;
		key_translation_table[::fltk::DivideKey-0xff00]   = KEY_Num_Div;
		key_translation_table[::fltk::Keypad0-0xff00] = KEY_Num_0;
		key_translation_table[::fltk::Keypad1-0xff00] = KEY_Num_1;
		key_translation_table[::fltk::Keypad2-0xff00] = KEY_Num_2;
		key_translation_table[::fltk::Keypad3-0xff00] = KEY_Num_3;
		key_translation_table[::fltk::Keypad4-0xff00] = KEY_Num_4;
		key_translation_table[::fltk::Keypad5-0xff00] = KEY_Num_5;
		key_translation_table[::fltk::Keypad6-0xff00] = KEY_Num_6;
		key_translation_table[::fltk::Keypad7-0xff00] = KEY_Num_7;
		key_translation_table[::fltk::Keypad8-0xff00] = KEY_Num_8;
		key_translation_table[::fltk::Keypad9-0xff00] = KEY_Num_9;
	}
	return key_translation_table;
}

key_event make_key_event(KeyAction a)
{
	char c = 0;
	if (::fltk::event_length() == 1)
		c = ::fltk::event_text()[0];

	unsigned short key = ::fltk::event_key();
	if (key == 32)
		key = KEY_Space;
	else if(key >= 0xff00)
		key = get_key_translation_table()[::fltk::event_key()-0xff00];
	else if (key < 256)
		key = cgv::media::text::to_upper((unsigned char)key);

	return key_event(key, a, c, make_modifiers(), make_toggle_keys(), ::fltk::get_time_secs());
}

mouse_event make_mouse_event(MouseAction a, int dx = 0, int dy = 0)
{
	unsigned char bs = 0;
	if ((::fltk::event_state()&::fltk::BUTTON1)!=0)
		bs |= MB_LEFT_BUTTON;
	if ((::fltk::event_state()&::fltk::BUTTON2)!=0)
		bs |= MB_MIDDLE_BUTTON;
	if ((::fltk::event_state()&::fltk::BUTTON3)!=0)
		bs |= MB_RIGHT_BUTTON;

	unsigned char b = 0;
	if (a == MA_PRESS || a == MA_RELEASE) {
		if (::fltk::event_button() == ::fltk::LeftButton)
			b = MB_LEFT_BUTTON;
		else if (::fltk::event_button() == ::fltk::MiddleButton)
			b = MB_MIDDLE_BUTTON;
		else if (::fltk::event_button() == ::fltk::RightButton)
			b = MB_RIGHT_BUTTON;
	}
	if (a == MA_WHEEL)
		dy = ::fltk::event_dy();
	return mouse_event(::fltk::event_x(), ::fltk::event_y(), a, bs, b, dx, dy, make_modifiers(), make_toggle_keys(), ::fltk::get_time_secs());
}

/// process focus and key press and release events here
int gl_view::handle(int ei)
{
	int dx = ::fltk::event_x() - last_mouse_x;
	int dy = ::fltk::event_y() - last_mouse_y;
	last_mouse_x = ::fltk::event_x();
	last_mouse_y = ::fltk::event_y();

	switch (ei) {
	case ::fltk::FOCUS:
	case ::fltk::UNFOCUS: 
		return 1;
	case ::fltk::KEY :
		if (traverser(make_action<event&>(make_key_event(KA_PRESS), &event_handler::handle)).traverse(group_ptr(this)))
			return 1;
		break;
	case ::fltk::KEYUP :
		if (traverser(make_action<event&>(make_key_event(KA_RELEASE), &event_handler::handle)).traverse(group_ptr(this)))
			return 1;
		break;
	case ::fltk::PUSH:
		if (traverser(make_action<event&>(make_mouse_event(MA_PRESS), &event_handler::handle)).traverse(group_ptr(this)))
			return 1;
		break;
	case ::fltk::RELEASE:
		if (traverser(make_action<event&>(make_mouse_event(MA_RELEASE), &event_handler::handle)).traverse(group_ptr(this)))
			return 1;
		break;
	case ::fltk::MOUSEWHEEL:
		if (traverser(make_action<event&>(make_mouse_event(MA_WHEEL), &event_handler::handle)).traverse(group_ptr(this)))
			return 1;
		break;
	case ::fltk::MOVE:
		if (traverser(make_action<event&>(make_mouse_event(MA_MOVE,dx,dy), &event_handler::handle)).traverse(group_ptr(this)))
			return 1;
		break;
	case ::fltk::DRAG:
		if (traverser(make_action<event&>(make_mouse_event(MA_DRAG,dx,dy), &event_handler::handle)).traverse(group_ptr(this)))
			return 1;
		break;
	case ::fltk::ENTER:
		take_focus();
		traverser(make_action<event&>(make_mouse_event(MA_ENTER), &event_handler::handle)).traverse(group_ptr(this));
		return 1;
	case ::fltk::LEAVE:
		if (traverser(make_action<event&>(make_mouse_event(MA_LEAVE), &event_handler::handle)).traverse(group_ptr(this)))
			return 1;
		break;
	}

	return ::fltk::GlWindow::handle(ei);       // hand on event to base class
}

		}
	}
}