#pragma once

#ifdef WIN32
#pragma warning(disable:4311)
#endif
#include <fltk/GlWindow.h>

#include <cgv/base/group.h>
#include <cgv/gui/event_handler.h>
#include <cgv/render/gl/context.h>
#include <cgv/gui/fltk/lib_begin.h>

namespace cgv {
	namespace gui {
		namespace fltk {

class CGV_API gl_view : 
	public ::fltk::GlWindow, 
	public cgv::base::group,
	public event_handler,
	public cgv::render::gl::context
{
public:
	/**@name fltk interface*/
	//@{
	/// construct application
	gl_view(int x, int y, int w, int h, const char* name = 0);
	/// complete drawing method that configures opengl whenever the context has changed, initializes the current frame and draws the scene
	void draw();
	/// process focus and key press and release events here
	int handle(int event);
	//@}

	/**@name base interface*/
	//@{
	/// overload to show the content of this object
	void stream_stats(std::ostream& os);
	//@}

	/**@name group interface*/
	//@{
	/// helper method to integrate a new child
	void configure_new_child(cgv::base::base_ptr child);
	/// helper method to remove a child
	void clear_child(cgv::base::base_ptr child);
	/// append child and return index of appended child
	unsigned int append_child(cgv::base::base_ptr child);
	/// remove all elements of the vector that point to child, return the number of removed children
	unsigned int remove_child(cgv::base::base_ptr child);
	/// remove all children
	void remove_all_children();
	/// insert a child at the given position
	void insert_child(unsigned int i, cgv::base::base_ptr child);
	/// put the event focus on the given child 
	bool set_focus(cgv::base::base_ptr child);
	//@}

	/**@name event_handler interface*/
	//@{
	/// overload and implement this method to handle events
	bool handle(event& e);
	/// overload to stream help information to the given output stream
	void stream_help(std::ostream& os);
	//@}

	/**@name context interface*/
	//@{
protected:
	/// information necessary for a rendering pass
	struct render_info
	{
		render::RenderPass pass;
		render::RenderPassFlags flags;
	};
	/// store the current render pass
	std::stack<render_info> render_pass_stack;
	/// current background color
	float bg_r, bg_g, bg_b, bg_a;
private:
	/// store current font size
	float current_font_size;
	/// store current font
	media::font::font_face_ptr current_font_face;
	/// remember the last active context to detect context changes
	void* last_context;
	/// remember the last size of the window to detect resize events
	unsigned int last_width, last_height;
protected:
	/// font used to draw textual info
	cgv::media::font::font_face_ptr info_font_face;
	/// font size to draw textual info
	float info_font_size;
	/// use a callback stream to write text to the opengl context
	cgv::media::text::callback_stream out_stream;
	/// size a tabs
	int tab_size;
	/// offset in x and y direction where text starts
	int x_offset, y_offset;
	/// callback method for processing of text from the output stream
	void process_text(const std::string& text);
private:
	/// current x and y position of text
	int text_x, text_y;
	/// current number of indentations
	int nr_identations;
	/// store whether we are at the beginning of the line
	bool at_line_begin;
	/// store whether we are in the draw method
	bool in_draw_method;
public:
	void destroy();
	/// return the width of the window
	unsigned int get_width() const;
	/// return the height of the window
	unsigned int get_height() const;
	/// set a user defined background color
	void set_bg_color(float r, float g, float b, float a);
	/// return the current render pass
	render::RenderPass get_render_pass() const;
	/// return the current render pass flags
	render::RenderPassFlags get_render_pass_flags() const;
	/// perform the given render task
	void render_pass(render::RenderPass render_pass, render::RenderPassFlags render_pass_flags, bool call_init_frame, bool call_finish_frame);
	/// return whether the context is currently in process of rendering
	bool in_render_process() const;
	/// return whether the context is current
	bool is_current() const;
	/// make the current context current
	void make_current() const;
	/// the context will be redrawn when the system is idle again
	void post_redraw();
	/// the context will be redrawn right now. This method cannot be called inside the following methods of a drawable: init, init_frame, draw, finish_draw
	void force_redraw();
	/// enable the given font face with the given size in pixels
	void enable_font_face(media::font::font_face_ptr font_face, float font_size);
	/// return the size in pixels of the currently enabled font face
	float get_current_font_size() const;
	/// return the currently enabled font face
	media::font::font_face_ptr get_current_font_face() const;
	/// overload to set text parameters
	void set_cursor(int x, int y);
	/// returns an output stream whose output is printed at the current cursor location
	std::ostream& output_stream();
	//@}
protected:
	/// 
	void draw_textual_info();
	/// define lighting mode, viewing pyramid and the rendering mode
	void configure_gl();
	/// clear frame buffer, set observer and light source locations
	void init_frame();
private:
	/// 
	bool show_help;
	bool show_stats;
	///
	int current_background;
	///
	bool do_screen_shot;
	///
	int last_mouse_x, last_mouse_y;

};

typedef cgv::data::ref_ptr<gl_view> gl_view_ptr;

		}
	}
}

#include <cgv/config/lib_end.h>