#pragma once

#include <iostream>
#include <cmath>
#include <cassert>
#include <cgv/math/vec.h>

namespace cgv {
namespace math {

/// A column vector class.
template <typename T>
class quat
{	

private:
	///pointer to _data storage
	T v[4];
	
public:

	quat()
	{
		v[0]    = 0;
		v[1]    = 0;
		v[2]    = 0;
		v[3]    = 1;
	}

	quat(const T& x,const T& y,const T& z,const T w)
	{
		v[0]    = x;
		v[1]    = y;
		v[2]    = z;
		v[3]    = w;
		
	}

	quat(const vec<T>& axis,const T angle)
	{
		set(axis,angle);
	}


	quat(const quat &q)
	{
		for (int i=0; i<4; i++)
			v[i]=q(i);
	}

	quat& operator=(const quat &rhs) 
	{
		if (this != &rhs)
		{
			for (int i=0; i<4; i++)
				v[i]=rhs(i);
		}
		return *this;
	}

	void conjugate()
    {
		v[0]    = -v[0];
		v[1]    = -v[1];
		v[2]    = -v[2];
    }

	quat<T>& operator*=(const quat<T>& q)
    {
		T w = v[3]*q(3) - v[0]*q(0) - v[1]*q(1) - v[2]*q(2);
		T x = v[3]*q(0) + v[0]*q(3) + v[1]*q(2) - v[2]*q(1);
		T y = v[3]*q(1) + v[1]*q(3) + v[2]*q(0) - v[0]*q(2);
		T z = v[3]*q(2) + v[2]*q(3) + v[0]*q(1) - v[1]*q(0);
		v[0]=x;
		v[1]=y;
		v[2]=z;
		v[3]=w;
		
		return *this;

    }

	const quat<T> operator*(const quat<T>& q)const
    {
		
		quat<T> r =*this;
		r*=q;
		return r;

    }


	T length()
	{
		T l =0;
		for(int i = 0; i < 4;i++)
			l += v[i]*v[i];
		return sqrt(l);
	}

	T sqr_length()
	{
		T l =0;
		for(int i = 0; i < 4;i++)
			l += v[i]*v[i];
		return l;
	}


	mat<T> to_mat_33() const
	{
		mat<T> mat(3,3);
		T xx      = v[0] * v[0];
		T xy      = v[0] * v[1];
		T xz      = v[0] * v[2];
		T xw      = v[0] * v[3];
		T yy      = v[1] * v[1];
		T yz      = v[1] * v[2];
		T yw      = v[1] * v[3];
		T zz      = v[2] * v[2];
		T zw      = v[2] * v[3];
		mat(0,0)  = 1 - 2 * ( yy + zz );
		mat(0,1)  =     2 * ( xy - zw );
		mat(0,2)  =     2 * ( xz + yw );
		mat(1,0)  =     2 * ( xy + zw );
		mat(1,1)  = 1 - 2 * ( xx + zz );
		mat(1,2)  =     2 * ( yz - xw );
		mat(2,0)  =     2 * ( xz - yw );
		mat(2,1)  =     2 * ( yz + xw );
		mat(2,2) = 1 - 2 * ( xx + yy );
		return mat;
	}

	mat<T> to_mat_44() const
	{
		mat<T> mat(4,4);
		T xx      = v[0] * v[0];
		T xy      = v[0] * v[1];
		T xz      = v[0] * v[2];
		T xw      = v[0] * v[3];
		T yy      = v[1] * v[1];
		T yz      = v[1] * v[2];
		T yw      = v[1] * v[3];
		T zz      = v[2] * v[2];
		T zw      = v[2] * v[3];
		mat(0,0)  = 1 - 2 * ( yy + zz );
		mat(0,1)  =     2 * ( xy - zw );
		mat(0,2)  =     2 * ( xz + yw );
		mat(1,0)  =     2 * ( xy + zw );
		mat(1,1)  = 1 - 2 * ( xx + zz );
		mat(1,2)  =     2 * ( yz - xw );
		mat(2,0)  =     2 * ( xz - yw );
		mat(2,1)  =     2 * ( yz + xw );
		mat(2,2) = 1 - 2 * ( xx + yy );
		mat(3,0) = mat(3,1) =mat(3,2)=mat(0,3)=mat(1,3)=mat(2,3) = 0;
		mat(3,3) = 1;
		return mat;
	}

	T& operator()(const int i)	//subscripting
	{
		return v[i];
	}
	
	
	const T&  operator()(const int i) const	//subscripting
	{
		return v[i];
	}

	///cast into non const array 
	operator T*()
	{
		return v;
	}

	///cast into const array
	operator const T*() const
	{
		return v;
	}

	///normalize the vector
	void normalize() 
	{
		T l =0;
		for(int i = 0; i < 4;i++)
			l += v[i]*v[i];
		l=(T)1.0/sqrt(l);
		for(unsigned i = 0; i<4; i++)
			 operator()(i)=l*operator()(i);
	}

	void set(const T& x, const T&y, const T&z,const T&w)
	{
		v[0]=x;
		v[1]=y;
		v[2]=z;
		v[3]=w;
		
	}

	
	void set(vec<T> axis, const T angle) 
	{
		axis.normalize();
		T sin_a = (T)sin( angle / (T)2.0 );
		T cos_a = (T)cos( angle / (T)2.0 );
		v[0]    = axis.x() * sin_a;
		v[1]    = axis.y() * sin_a;
		v[2]    = axis.z() * sin_a;
		v[3]    = cos_a;

	}

	void invert()
	{
		
		conjugate();
		normalize();
		
	}

	void rotate(vec<T>& pos) const
	{
		
		T w =  - v[0]*pos(0) - v[1]*pos(1) - v[2]*pos(2);
		T x = v[3]*pos(0)  + v[1]*pos(2) - v[2]*pos(1);
		T y = v[3]*pos(1)  + v[2]*pos(0) - v[0]*pos(2);
		T z = v[3]*pos(2)  + v[0]*pos(1) - v[1]*pos(0);
	
		pos(0) = -w*v[0] + x*v[3] - y*v[2] + z*v[1];
		pos(1) = -w*v[1] + y*v[3] - z*v[0] + x*v[2];
		pos(2) = -w*v[2] + z*v[3] - x*v[1] + y*v[0];
		
	}

};

template<typename T>
vec<T> rotate(const quat<T>& q,const vec<T>& pos)
{
	vec<T> v = pos;
	q.rotate(v);
	return v;
}

template<typename T>
quat<T> normalize(const quat<T>& q)
{
	quat<T> r = v;
	r.normalize();
	return r;
}

//output of a vector
template<typename T>
std::ostream& operator<<(std::ostream& out, const quat<T>& q)
{

	for (unsigned i=0;i<3;++i)
	{
		out << q(i)<<"\t";	
	}
	out << q(3);
	return out;

}

//input of a vector
template<typename T>
std::istream& operator>>(std::istream& in, quat<T>& q)
{

	for (unsigned i=0;i<4;++i)
	{
		in >> q(i);	
	}
	
	return in;

}




}

}
