#pragma once
#include <cgv/math/vec.h>
#include <cgv/math/mat.h>

namespace cgv {
namespace math {
/*
* A sparse matrix in triplet or column compressed form.
*/
template <typename T>
class sparse_mat
{
private:
	//maximum number of entries
	int _nzmax;
	//used number of entries in uncompressed matrix -1 if compressed
	int _nz;
	//number of rows
	int _nrows;//m
	//number of columns
	int _ncols; //n
	//column indices or column pointers in compressed form 
	vec<int> _p;
	//row indices
	vec<int> _i;
	//stored matrix entries
	vec<T> _data;//x


public:
	sparse_mat(long nrows, long ncols, long nzmax,bool compressed=false) :
	  _nrows(nrows),_ncols(ncols),_data(nzmax),_i(nzmax)
	{
		if(compressed)
		{
			_p.resize(ncols+1);
			_nz=-1;
		}else
		{
			_p.resize(nzmax);
			_nz=0;
		}
		
	}

	const int& p(int ix) const
	{
		assert(is_compressed());
		return _p[ix];
	}

	int& p(int ix)
	{
		assert(is_compressed());
		return _p[ix];
	}


	const int& j(int ix) const
	{
		assert(!is_compressed());
		return _p[ix];
	}

	int& j(int ix)
	{
		assert(!is_compressed());
		return _p[ix];
	}


	const int& i(int ix) const
	{
		return _i[ix];
	}

	int& i(int ix)
	{
		return _i[ix];
	}

	
	int nrows() const
	{
		return _nrows;
	}

	int ncols() const
	{
		return _ncols;
	}

	int nnz() const
	{
		if(is_compressed())
			return _p[_ncols];
		else
			return _nz;
	}

	int nzmax() const
	{
		return _nzmax;
	}
	

	bool is_compressed() const
	{
		return _nz == -1;
	}

	void compress()
	{
		if(is_compressed())
			return;
		//compress matrix
		
		sparse_mat<T> C(_nrows, _ncols, _nz) ;		
    	vec<int> w;
		long p;
		w.zeros(_ncols);

	    for (int k = 0 ; k < _nz ; k++) 
			w (_p[k])++ ;	// column counts 

		cumsum_values( w,C._p);		
		w=C._p;
		for (int k = 0 ; k < _nz ; k++)
		{
			C._i [p = w [_p[k]]++] = _i[k] ;    /* A(i,j) is the pth entry in C */
			C._data[p] = _data[k] ;
		}
		_nz =-1;
    

	}

	void uncompress()
	{
		if(!is_compressed())
			return;
		//uncompress matrix
		assert(false);
		//not yet implemented

	}

	///transpose matrix
	void transpose()
	{
		assert(false);
		//not yet implemented
	}

	//add an entry to uncompressed matrix
	void add_entry(int i, int j,const T& x)
	{
		assert(is_compressed() == false);
		assert(i >= 0 && i < _nrows && j >= 0 && j <_ncols);
		
		_data[_nz] = x;
		_i [_nz] = i;
		_p[_nz++] = j;
	}

	const T& operator()(unsigned k) const
	{
		
		return _data[k];
	}
	T& operator()(unsigned k)
	{
		
		return _data[k];
	}

	vec<T> operator*(const vec<T>& v) const
	{
		assert(false);
		//not yet implemented
		return v;
	}

	///1-norm of a sparse matrix = max (sum (abs (A))), largest column sum 
	T norm() const
	{
		assert(is_compressed() == true);

		T  norm = 0,s;
		int p,j;
    
		for (j = 0 ; j < _nrows ; j++)
		{
			for (s = 0, p = _p [j] ; p < _p [j+1] ; p++)
				s += std::abs (_data[p]) ;
			norm = std::max(norm, s) ;
		}
		return norm;
	}
protected:
	/*///p [0..n] = cumulative sum of c [0..n-1], and then copy p [0..n-1] into c 
	T cumulative_sum (const vec<T>& p,const vec<T>& c, int n)
	{
		assert(p.size() > n && c.size() >=n);
	    int i, nz = 0 ;
		T nz2 = 0 ;
		p.resize(c.size());
		for (i = 0 ; i < c.size() ; i++)
		{
			p[i] = nz ;
			nz += c[i] ;
			nz2 += c[i] ;		    // also in T to avoid int overflow 
			c[i] = p[i] ;		    // also copy p[0..n-1] back into c[0..n-1]
		}
		p[n] = nz ;
		return (nz2) ;		    // return sum (c [0..n-1]) 
	}*/

};

///output of a matrix onto an ostream
template <typename T>
std::ostream& operator<<(std::ostream& out, const sparse_mat<T>& m)
{
	
	if(m.is_compressed())
	{
		out << "compressed: ";
		out << m.nrows()<<"x"<<m.ncols()<<" nzmax: "<< m.nzmax()<<", nnz: "<<m.nnz()<<std::endl;
		for (int k = 0 ; k < m.nnz() ; k++)
		{
			out<<m.i(k) <<" " <<m.p(k)<<": "<<m(k)<<std::endl;		
		}

	}
	else
	{
		out << "triplet: ";
		out << m.nrows()<<"x"<<m.ncols()<<" nzmax: "<< m.nzmax()<<", nnz: "<<m.nnz()<<std::endl;
		for (int k = 0 ; k < m.nnz() ; k++)
		{
			out<<m.i(k) <<" "<< m.j(k)<<": "<<m(k)<<std::endl;		
		}
	}


	return out;

}


}
}

