#include <cgv/render/context.h>

namespace cgv {
	namespace render {

// declare some colors by name
float black[4]     = { 0, 0, 0, 1 };
float white[4]     = { 1, 1, 1, 1 };
float gray[4]      = { 0.25f, 0.25f, 0.25f, 1 };
float green[4]     = { 0, 1, 0, 1 };
float brown[4]     = { 0.3f, 0.1f, 0, 1 };
float dark_red[4]  = { 0.4f, 0, 0, 1 };
float cyan[4]      = { 0, 1, 1, 1 };
float yellow[4]    = { 1, 1, 0, 1 };
float red[4]       = { 1, 0, 0, 1 };
float blue[4]      = { 0, 0, 1, 1 };

void normalize(double* n)
{
	double in = 1.0/sqrt(n[0]*n[0]+n[1]*n[1]+n[2]*n[2]);
	n[0] *= in;
	n[1] *= in;
	n[2] *= in;
}

void compute_normal(double* n, const double* v0, const double* v1, const double* v2)
{
	n[0] = (v1[1]-v0[1])*(v2[2]-v0[2])-(v1[2]-v0[2])*(v2[1]-v0[1]);
	n[1] = (v1[2]-v0[2])*(v2[0]-v0[0])-(v1[0]-v0[0])*(v2[2]-v0[2]);
	n[2] = (v1[0]-v0[0])*(v2[1]-v0[1])-(v1[1]-v0[1])*(v2[0]-v0[0]);
	normalize(n);
}

void compute_face_normals(const double* vertices, double* normals, const int* vertex_indices, int* normal_indices, int nr_faces, int face_degree)
{
	for (int i = 0; i < nr_faces; ++i) {
		compute_normal(normals+3*i, 
					   vertices+3*vertex_indices[face_degree*i],
					   vertices+3*vertex_indices[face_degree*i+1],
					   vertices+3*vertex_indices[face_degree*i+2]);
		for (int j = 0; j<face_degree; ++j)
			normal_indices[face_degree*i+j] = i;
	}
}

/// tesselate a unit cube with extent from [-1,-1,-1] to [1,1,1] with face normals that can be flipped
void context::tesselate_unit_cube(bool flip_normals)
{
	static double V[8*3] = {
		-1,-1,+1,
		+1,-1,+1,
		-1,+1,+1,
		+1,+1,+1,
		-1,-1,-1,
		+1,-1,-1,
		-1,+1,-1,
		+1,+1,-1
	};
	static double N[6*3] = {
		-1,0,0, +1,0,0, 
		0,-1,0, 0,+1,0,
		0,0,-1, 0,0,+1
	};
	static double T[14*2] = { 
		 0,1.0/3 , 0,2.0/3 ,
		 0.25,0 , 0.25,1.0/3 ,
		 0.25,2.0/3 , 0.25,1 ,
		 0.5,0 , 0.5,1.0/3 ,
		 0.5,2.0/3 , 0.5,1 ,
		 0.75,1.0/3 , 0.75,2.0/3 ,
		 1,1.0/3 , 1,2.0/3 
	};
	static int F[6*4] = {
		0,2,6,4,
		1,5,7,3,
		0,4,5,1,
		2,3,7,6,
		4,6,7,5,
		0,1,3,2 
	};
	static int FN[6*4] = {
		0,0,0,0, 1,1,1,1,
		2,2,2,2, 3,3,3,3,
		4,4,4,4, 5,5,5,5
	};
	static int FT[6*4] = {
		3,4,1,0 ,7,10,11,8 ,
		3,2,6,7 ,4,8,9,5 ,
		12,13,11,10 ,3,7,8,4 
	};
	draw_faces(V,N,T,F,FN,FT,6,4, flip_normals);
}

/// tesselate a prism 
void context::tesselate_unit_prism(bool flip_normals)
{
	static const double V[6*3] = {
		-1, -1, -1,
		 1, -1, -1,
		 0, -1,  1,
		-1,  1, -1,
		 1,  1, -1,
		 0,  1,  1
	};
	static double a = 1.0/sqrt(5.0);
	static double b = 2*a;
	static const double N[5*3] = {
		 0,-1, 0,
		 0, 1, 0,
		 0, 0,-1,
		-b, 0, a, 
		 b, 0, a
	};
	static const int FT[2*3] = { 0,1,2,	5,4,3 };
	static const int FQ[8] = { 4,1, 3,0, 5,2, 4,1};
	static const int FTN[2*3] = { 0,0,0, 1,1,1 };
	static const int FQN[8] = { 2,2, 2,2, 3,3, 4,4, };

	draw_faces(V, N, 0, FT, FTN, 0, 2, 3, flip_normals);
	draw_strip_or_fan(V, N, 0, FQ, FQN, 0, 3, 4, flip_normals);
}

/// tesselate a circular disk of radius 1
void context::tesselate_unit_disk(int resolution, bool flip_normals)
{
	std::vector<double> V; V.reserve(3*(resolution+1));
	std::vector<double> N; N.reserve(3*(resolution+1));
	std::vector<double> T; T.reserve(2*(resolution+1));

	std::vector<int> F; F.resize(resolution+1);
	int i;
	for (i = 0; i <= resolution; ++i)
		F[i] = i;

	double step = 2*M_PI/resolution;
	double phi   = 0;
	for (i = 0; i <= resolution; ++i, phi += step) {
		double cp = cos(phi); 
		double sp = sin(phi);
		N.push_back(0);
		N.push_back(0);
		N.push_back(1);
		T.push_back((double)i/resolution);
		T.push_back(1);
		V.push_back(cp);
		V.push_back(sp);
		V.push_back(0);
	}
	draw_faces(&V[0],&N[0],&T[0],&F[0],&F[0],&F[0],1,resolution+1,flip_normals);
}

/// tesselate a cone of radius 1
void context::tesselate_unit_cone(int resolution, bool flip_normals)
{
	std::vector<double> V; V.reserve(6*(resolution+1));
	std::vector<double> N; N.reserve(6*(resolution+1));
	std::vector<double> T; T.reserve(4*(resolution+1));

	std::vector<int> F; F.resize(2*resolution+2);
	int i;
	for (i = 0; i <= 2*resolution+1; ++i)
		F[i] = i;

	static double a = 1.0/sqrt(5.0);
	static double b = 2*a;
	double step = 2*M_PI/resolution;
	double phi   = 0;
	double u = 0;
	double duv = 1.0/resolution;
	for (int i = 0; i <= resolution; ++i, u += duv, phi += step) {
		double cp = cos(phi); 
		double sp = sin(phi);
		N.push_back(b*cp);
		N.push_back(b*sp);
		N.push_back(a);
		T.push_back(u);
		T.push_back(1);
		V.push_back(0);
		V.push_back(0);
		V.push_back(1);
		N.push_back(b*cp);
		N.push_back(b*sp);
		N.push_back(a);
		T.push_back(u);
		T.push_back(0);
		V.push_back(cp);
		V.push_back(sp);
		V.push_back(-1);
	}
	draw_strip_or_fan(&V[0],&N[0],&T[0],&F[0],&F[0],&F[0],resolution,4,false,flip_normals);
}

/// tesselate a cylinder of radius 1
void context::tesselate_unit_cylinder(int resolution, bool flip_normals)
{
	std::vector<double> V; V.reserve(6*(resolution+1));
	std::vector<double> N; N.reserve(6*(resolution+1));
	std::vector<double> T; T.reserve(4*(resolution+1));

	std::vector<int> F; F.resize(2*(resolution+1));
	int i;
	for (i = 0; i <= 2*resolution+1; ++i)
		F[i] = i;

	double step = 2*M_PI/resolution;
	double phi   = 0;
	double u = 0;
	double duv = 1.0/resolution;
	for (int i = 0; i <= resolution; ++i, u += duv, phi += step) {
		double cp = cos(phi); 
		double sp = sin(phi);
		N.push_back(cp);
		N.push_back(sp);
		N.push_back(0);
		T.push_back(u);
		T.push_back(1);
		V.push_back(cp);
		V.push_back(sp);
		V.push_back(1);
		N.push_back(cp);
		N.push_back(sp);
		N.push_back(0);
		T.push_back(u);
		T.push_back(0);
		V.push_back(cp);
		V.push_back(sp);
		V.push_back(-1);
	}
	draw_strip_or_fan(&V[0],&N[0],&T[0],&F[0],&F[0],&F[0],resolution,4,false,flip_normals);
}

/// tesselate a sphere of radius 1
void context::tesselate_unit_sphere(int resolution, bool flip_normals)
{
	std::vector<double> V; V.resize(6*(resolution+1));
	std::vector<double> N; N.resize(6*(resolution+1));
	std::vector<double> T; T.resize(4*(resolution+1));
	std::vector<int> F; F.resize(2*(resolution+1));
	int i;
	for (int i = 0; i <= resolution; ++i) {
		F[2*i] = 2*i;
		F[2*i+1] = 2*i+1;
	}
	double step = M_PI/resolution;
	double phi   = 0;
	double cp1   = 1, sp1 = 0;
	double u = 0;
	double duv = 1.0/resolution;
	for (i = 0; i < resolution; ++i, u += duv) {
		double cp0 = cp1, sp0 = sp1;
		phi += 2*step;
		cp1 = cos(phi); 
		sp1 = sin(phi);
		double theta = -0.5*M_PI;
		double v = 0;
		int kv=0, kn=0, kt=0;
		for (int j = 0; j <= resolution; ++j, theta += step, v += duv) {
			double ct = cos(theta), st = sin(theta);
			N[kn++] = ct*cp0;
			N[kn++] = ct*sp0;
			N[kn++] = st;
			T[kt++] = u;
			T[kt++] = v;
			V[kv++] = ct*cp0;
			V[kv++] = ct*sp0;
			V[kv++] = st;
			N[kn++] = ct*cp1;
			N[kn++] = ct*sp1;
			N[kn++] = st;
			T[kt++] = u+duv;
			T[kt++] = v;
			V[kv++] = ct*cp1;
			V[kv++] = ct*sp1;
			V[kv++] = st;
		}
		draw_strip_or_fan(&V[0],&N[0],&T[0],&F[0],&F[0],&F[0],resolution,4,false, flip_normals);
	}

}
/// tesselate a tetrahedron
void context::tesselate_unit_tetrahedron(bool flip_normals)
{
	static const double a = 1.0/(2*sqrt(3.0));
	static const double b = 1.0/(3*sqrt(3.0/2));
	static const double V[4*3] = {
		 -0.5, -a, -b,
		  0.5, -a, -b,
		    0,2*a, -b,
		    0,  0,2*b
	};
	static const int F[4*3] = {
		0,2,1,3,2,0,3,0,1,3,1,2
	};
	static int FN[4*3];
	static double N[4*3];
	static bool computed = false;
	if (!computed) {
		compute_face_normals(V, N, F, FN, 4, 3);
		computed = true;
	}
	draw_faces(V, N, 0, F, FN, 0, 4, 3, flip_normals);
}

/// tesselate a octahedron
void context::tesselate_unit_octahedron(bool flip_normals)
{
	static double N[8*3] = {
		-1,-1,+1, +1,-1,+1, -1,+1,+1, +1,+1,+1,
		-1,-1,-1, +1,-1,-1, -1,+1,-1, +1,+1,-1
	};
	static double V[6*3] = {
		-1,0,0, +1,0,0,
		0,-1,0, 0,+1,0,
		0,0,-1, 0,0,+1
	};
	static int FN[8*3] = {
		0,0,0,
		1,1,1,
		2,2,2,
		3,3,3,
		4,4,4,
		5,5,5,
		6,6,6,
		7,7,7
	};
	static int F[8*3] = {
		0,2,5,
		1,5,2,
		5,3,0,
		3,5,1,
		4,2,0,
		4,1,2,
		3,4,0,
		1,4,3
	};
	draw_faces(V, N, 0, F, FN, 0, 8, 3, flip_normals);
}

/// render an icosahedron at a given center.
void tesselate_unit_dodecahedron_or_icosahedron(context& c, bool dual, bool flip_normals)
{
	static const double h = 0.4472135956;
	static const double r = 0.8944271912;
	static const double s = M_PI/2.5;
	static const double o = M_PI/5;
	static const double V[13*3] = {
		0,0,-1,
		r*sin(1*s),r*cos(1*s),-h,
		r*sin(2*s),r*cos(2*s),-h,
		r*sin(3*s),r*cos(3*s),-h,
		r*sin(4*s),r*cos(4*s),-h,
		r*sin(5*s),r*cos(5*s),-h,
		r*sin(1*s+o),r*cos(1*s+o),h,
		r*sin(2*s+o),r*cos(2*s+o),h,
		r*sin(3*s+o),r*cos(3*s+o),h,
		r*sin(4*s+o),r*cos(4*s+o),h,
		r*sin(5*s+o),r*cos(5*s+o),h,
		0,0,1,
		0,0,0
	};
	static int F[20*3] = {
		0,1,2, 0,2,3, 0,3,4, 0,4,5, 0,5,1,
		6,2,1, 2,6,7, 7,3,2, 3,7,8, 8,4,3, 4,8,9, 9,5,4, 5,9,10, 10,1,5, 6,1,10,
		11,6,10, 11,10,9, 11,9,8, 11,8,7, 11,7,6
	};
	static int DF[12*5] = {
		0,1,2,3,4,
		5,6,7,1,0,
		7,8,9,2,1,
		9,10,11,3,2,
		11,12,13,4,3,
		13,14,5,0,4,
		16,15,14,13,12,
		17,16,12,11,10,
		18,17,10,9,8,
		19,18,8,7,6,
		15,19,6,5,14,
		15,16,17,18,19
	};
	static double N[20*3];
	static int FN[20*3];
	static int DFN[12*5] = {
		0,0,0,0,0,
		2,2,2,2,2,
		3,3,3,3,3,
		4,4,4,4,4,
		5,5,5,5,5,
		1,1,1,1,1,
		10,10,10,10,10,
		9,9,9,9,9,
		8,8,8,8,8,
		7,7,7,7,7,
		6,6,6,6,6,
		11,11,11,11,11
	};

	static bool computed = false;
	if (!computed) {
		compute_face_normals(V, N, F, FN, 20, 3);
		computed = true;
	}
	if (!dual)
		c.draw_faces(V, N, 0, F, FN, 0, 20, 3, flip_normals);
	else
		c.draw_faces(N, V, 0, DF, DFN, 0, 12, 5, flip_normals);
}

/// tesselate a dodecahedron
void context::tesselate_unit_dodecahedron(bool flip_normals)
{
	tesselate_unit_dodecahedron_or_icosahedron(*this, true, flip_normals);
}
/// tesselate an icosahedron
void context::tesselate_unit_icosahedron(bool flip_normals)
{
	tesselate_unit_dodecahedron_or_icosahedron(*this, false, flip_normals);
}

/// return homogeneous 4x4 matrix, which transforms from world to device space
const context::mat_type context::get_DPV() const
{
	return get_D()*get_P()*get_V();
}

/// compute a the location in world space of a device x/y-location. For this the device point is extended with the device z-coordinate currently stored in depth buffer.
const context::vec_type context::get_point_W(int x_D, int y_D) const
{
	return get_point_W(x_D, y_D, get_z_D(x_D,y_D));
}
const context::vec_type context::get_point_W(int x_D, int y_D, const mat_type& DPV) const
{
	return get_point_W(x_D, y_D, get_z_D(x_D,y_D), DPV);
}
/// compute a the location in world space of a device point.
const context::vec_type context::get_point_W(int x_D, int y_D, double z_D) const
{
	return get_point_W(x_D, y_D, z_D, get_DPV());
}

const context::vec_type context::get_point_W(int x_D, int y_D, double z_D, const mat_type& MPD) const
{
	vec_type x(4);
	vec_type b(4);
	b(0) = x_D;
	b(1) = y_D;
	b(2) = z_D;
	b(3) = 1;
	solve(MPD,b,x);
	vec_type p(3);
	p(0) = x(0)/x(3);
	p(1) = x(1)/x(3);
	p(2) = x(2)/x(3);
	return p;
}


	}
}
