#pragma once

#include <vector>
#include <string>
#include <cgv/media/font/font.h>
#define _USE_MATH_DEFINES
#include <cgv/math/inv.h>

#include <cgv/render/lib_begin.h>

namespace cgv {
	namespace render {

/// enumeration of rendering APIs which can be queried from the context
enum RenderAPI {
	RA_OPENGL,
	RA_DIRECTX
};

/// enumeration of different render passes, which can be queried from the context and used to specify a new render pass
enum RenderPass {
	RP_NONE,
	RP_SHADOW_MAP,           /// construction of shadow map
	RP_SHADOW_VOLUME,        /// construction of shadow map
	RP_OPAQUE_SURFACES,      /// opaque surface rendering using z-Buffer
	RP_TRANSPARENT_SURFACES, /// transparent surface rendering using depth peeling
	RP_PICK,                 /// in picking pass a small rectangle around the mouse is rendered 
};

/// available flags that can be queried from the context and set for a new render pass
enum RenderPassFlags {
	RPF_NONE = 0,
	RPF_PROJECTION_SET = 1,
	RPF_MODELVIEW_SET = 2,
	RPF_MODELVIEW_PROJECTION_SET = 3,
	RPF_LIGHTS_SET = 4,
	RPF_NO_LIGHTING = 8
};

// declare some colors by name
extern CGV_API float black[4], white[4], gray[4], green[4], brown[4], dark_red[4];
extern CGV_API float cyan[4], yellow[4], red[4], blue[4];

/** base class for all drawables, which is independent of the used rendering API. */
class CGV_API context 
{
public:
	/**@name render process*/
	//@{
	/// return the used rendering API
	virtual RenderAPI get_render_api() const = 0;
	/// return the current render pass
	virtual RenderPass get_render_pass() const = 0;
	/// return the current render pass flags
	virtual RenderPassFlags get_render_pass_flags() const = 0;
	/// perform the given render task
	virtual void render_pass(RenderPass render_pass, RenderPassFlags render_pass_flags, bool call_init_frame = false, bool call_finish_frame = false) = 0;
	/// return whether the context is currently in process of rendering
	virtual bool in_render_process() const = 0;
	/// return whether the context is current
	virtual bool is_current() const = 0;
	/// make the current context current
	virtual void make_current() const = 0;
	//@}

	/// return the width of the window
	virtual unsigned int get_width() const = 0;
	/// return the height of the window
	virtual unsigned int get_height() const = 0;
	/// set a user defined background color
	virtual void set_bg_color(float r, float g, float b, float a) = 0;
	/// the context will be redrawn when the system is idle again
	virtual void post_redraw() = 0;
	/// the context will be redrawn right now. This method cannot be called inside the following methods of a drawable: init, init_frame, draw, finish_draw
	virtual void force_redraw() = 0;

	/**@name font selection and measure*/
	//@{
	/// enable the given font face with the given size in pixels
	virtual void enable_font_face(media::font::font_face_ptr font_face, float font_size) = 0;
	/// return the size in pixels of the currently enabled font face
	virtual float get_current_font_size() const = 0;
	/// return the currently enabled font face
	virtual media::font::font_face_ptr get_current_font_face() const = 0;
	//@}

	/**@name text output*/
	//@{
	/// use this to push transformation matrices on the stack such that x and y coordinates correspond to window coordinates
	virtual void push_pixel_coords() = 0;
	/// set a new cursor position, which is only valid between calls of push_pixel_coords and pop_pixel_coords
	virtual void set_cursor(int x, int y) = 0;
	/// store the current cursor location in the passed references to x and y coordinate
	virtual void get_cursor(int& x, int& y) const = 0;
	/// pop previously changed transformation matrices 
	virtual void pop_pixel_coords() = 0;
	/// returns an output stream whose output is printed at the current cursor location
	virtual std::ostream& output_stream() = 0;
	//@}

	/**@name tesselation*/
	//@{
	/// tesselate with independent faces
	virtual void draw_faces(
		const double* vertices, const double* normals, const double* tex_coords, 
		const int* vertex_indices, const int* normal_indices, const int* tex_coord_indices, 
		int nr_faces, int face_degree, bool flip_normals = false) const = 0;
	/// tesselate with one face strip
	virtual void draw_strip_or_fan(
		const double* vertices, const double* normals, const double* tex_coords, 
		const int* vertex_indices, const int* normal_indices, const int* tex_coord_indices, 
		int nr_faces, int face_degree, bool is_fan, bool flip_normals = false) const = 0;

	/// tesselate a unit cube with extent from [-1,-1,-1] to [1,1,1] with face normals that can be flipped
	void tesselate_unit_cube(bool flip_normals = false);
	/// tesselate a prism 
	void tesselate_unit_prism(bool flip_normals = false);
	/// tesselate a circular disk of radius 1
	void tesselate_unit_disk(int resolution = 25, bool flip_normals = false);
	/// tesselate a cone of radius 1
	void tesselate_unit_cone(int resolution = 25, bool flip_normals = false);
	/// tesselate a cylinder of radius 1
	void tesselate_unit_cylinder(int resolution = 25, bool flip_normals = false);
	/// tesselate a sphere of radius 1
	void tesselate_unit_sphere(int resolution = 25, bool flip_normals = false);
	/// tesselate a tetrahedron
	void tesselate_unit_tetrahedron(bool flip_normals = false);
	/// tesselate a octahedron
	void tesselate_unit_octahedron(bool flip_normals = false);
	/// tesselate a dodecahedron
	void tesselate_unit_dodecahedron(bool flip_normals = false);
	/// tesselate an icosahedron
	void tesselate_unit_icosahedron(bool flip_normals = false);
	//@}

	/**@name transformations*/
	//@{
	/// declare type of matrices
	typedef cgv::math::mat<double> mat_type;
	/// declare type of vectors
	typedef cgv::math::vec<double> vec_type;
	/// return homogeneous 4x4 viewing matrix, which transforms from world to eye space
	virtual const mat_type get_V() const = 0;
	/// return homogeneous 4x4 projection matrix, which transforms from eye to clip space
	virtual const mat_type get_P() const = 0;
	/// return homogeneous 4x4 projection matrix, which transforms from clip to device space
	virtual const mat_type get_D() const = 0;
	/// return homogeneous 4x4 matrix, which transforms from world to device space
	virtual const mat_type get_DPV() const;
	/// read the device z-coordinate from the z-buffer for the given device x- and y-coordinates
	virtual double get_z_D(int x_D, int y_D) const = 0;
	/// compute the location in world space of a device x/y-location. For this the device point is extended with the device z-coordinate currently stored in the displayed depth buffer.
	virtual const vec_type get_point_W(int x_D, int y_D) const;
	/// compute the location in world space of a device x/y-location by inversion of the given transformation from world to device space. For this the device point is extended with the device z-coordinate currently stored in the displayed depth buffer.
	virtual const vec_type get_point_W(int x_D, int y_D, const mat_type& DPV) const;
	/// compute the location in world space of a device point. For this the current world to device transformation get_DPV is inverted.
	virtual const vec_type get_point_W(int x_D, int y_D, double z_D) const;
	/// compute the location in world space of a device point by inversion of the given world to device transformation.
	virtual const vec_type get_point_W(int x_D, int y_D, double z_D, const mat_type& DPV) const;
	//@}
};
	}
}

#include <cgv/config/lib_end.h>