#include <cgv/render/gl/context.h>
#ifdef WIN32
#include <windows.h>
#endif
#include <gl/gl.h>
#include <gl/glu.h>
namespace cgv {
	namespace render {
		namespace gl {

/// construct context and attach signals
context::context()
{
}
/// return the used rendering API
RenderAPI context::get_render_api() const
{
	return RA_OPENGL;
}
/// use this to push transformation matrices on the stack such that x and y coordinates correspond to window coordinates
void context::push_pixel_coords()
{
	// push projection matrix
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
		// set orthogonal projection
		glLoadIdentity();
		GLint vp[4];
		glGetIntegerv(GL_VIEWPORT, vp);
		gluOrtho2D(0,vp[2],vp[3],0);
		// push modelview matrix
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
			// use identity for modelview
			glLoadIdentity();
}
/// set a new cursor position, which is only valid between calls of push_pixel_coords and pop_pixel_coords
void context::set_cursor(int x, int y)
{
	glRasterPos2i(x,y);
}

/// store the current cursor location in the passed references to x and y coordinate
void context::get_cursor(int& x, int& y) const
{
	GLint r[4];
	glGetIntegerv(GL_CURRENT_RASTER_POSITION, r);
	x = r[0];
	y = get_height()-r[1];
}


/// pop previously changed transformation matrices 
void context::pop_pixel_coords()
{
		// restore modelview matrix
		glPopMatrix();
		glMatrixMode(GL_PROJECTION);
	// restore projection matrix
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
}

void render_vertex(int k, const double* vertices, const double* normals, const double* tex_coords, 
						  const int* vertex_indices, const int* normal_indices, const int* tex_coord_indices, bool flip_normals)
{
	if (normals && normal_indices) {
		if (flip_normals) {
			double n[3] = { -normals[3*normal_indices[k]],-normals[3*normal_indices[k]+1],-normals[3*normal_indices[k]+2] };
			glNormal3dv(n);
		}
		else
			glNormal3dv(normals+3*normal_indices[k]);
	}
	if (tex_coords && tex_coord_indices)
		glTexCoord2dv(tex_coords+2*tex_coord_indices[k]);
	glVertex3dv(vertices+3*vertex_indices[k]);
}


void context::draw_faces(
	const double* vertices, const double* normals, const double* tex_coords, 
	const int* vertex_indices, const int* normal_indices, const int* tex_coord_indices, 
	int nr_faces, int face_degree, bool flip_normals) const
{
	int k = 0;
	if (face_degree < 5)
		glBegin( face_degree == 3 ? GL_TRIANGLES : GL_QUADS);
	for (int i = 0; i<nr_faces; ++i) {
		if (face_degree >= 5)
			glBegin(GL_POLYGON);
		for (int j=0; j<face_degree; ++j, ++k)
			render_vertex(k, vertices, normals, tex_coords, vertex_indices, normal_indices,tex_coord_indices,flip_normals);
		if (face_degree >= 5)
			glEnd();		
	}
	if (face_degree < 5)
		glEnd();
}

void context::draw_strip_or_fan(
		const double* vertices, const double* normals, const double* tex_coords, 
		const int* vertex_indices, const int* normal_indices, const int* tex_coord_indices, 
		int nr_faces, int face_degree, bool is_fan, bool flip_normals) const
{
	glBegin( face_degree == 3 ? (is_fan ? GL_TRIANGLE_FAN : GL_TRIANGLE_STRIP) : GL_QUAD_STRIP);
	render_vertex(0, vertices, normals, tex_coords, vertex_indices, normal_indices,tex_coord_indices,flip_normals);
	render_vertex(1, vertices, normals, tex_coords, vertex_indices, normal_indices,tex_coord_indices,flip_normals);
	int s = face_degree-2;
	int k = 2;
	for (int i = 0; i<nr_faces; ++i)
		for (int j=0; j<s; ++j, ++k)
			render_vertex(k, vertices, normals, tex_coords, vertex_indices, normal_indices,tex_coord_indices,flip_normals);
	glEnd();
}	

/// return homogeneous 4x4 viewing matrix, which transforms from world to eye space
const context::mat_type context::get_V() const
{
	GLdouble V[16];
	glGetDoublev(GL_MODELVIEW_MATRIX, V);
	return mat_type(4,4,V);
}

/// return homogeneous 4x4 projection matrix, which transforms from eye to clip space
const context::mat_type context::get_P() const
{
	GLdouble P[16];
	glGetDoublev(GL_PROJECTION_MATRIX, P);
	return mat_type(4,4,P);
}

/// return homogeneous 4x4 projection matrix, which transforms from clip to device space
const context::mat_type context::get_D() const
{
	mat_type D(4,4,0.0);
	D(0,0) =  0.5*get_width();
	D(0,3) =  0.5*get_width();
	D(1,1) = -0.5*get_height(); // flip y-coordinate
	D(1,3) =  0.5*get_height();
	D(2,2) =  0.5;
	D(2,3) =  0.5;
	D(3,3) =  1.0;
	return D;
}

/// read the device z-coordinate from the z-buffer for the given device x- and y-coordinates
double context::get_z_D(int x_D, int y_D) const
{
	GLfloat z_D;

	if (!in_render_process() && !is_current())
		make_current();

	glReadPixels(x_D, get_height()-y_D, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &z_D);
	/*
	GLenum err = glGetError();
	if (err != GL_NO_ERROR) {
		std::cout << "gl error:";
		switch (err) {
			case GL_INVALID_ENUM :      std::cout << "invalid enum"; break;
			case GL_INVALID_VALUE :     std::cout << "invalid value"; break;
			case GL_INVALID_OPERATION : std::cout << "invalid operation"; break;
			case GL_STACK_OVERFLOW :    std::cout << "stack overflow"; break;
			case GL_STACK_UNDERFLOW :   std::cout << "stack underflow"; break;
			case GL_OUT_OF_MEMORY :     std::cout << "out of memory"; break;
			default:                    std::cout << "unknown error"; break;
		}
		std::cout << std::endl;

	}
	*/
	return z_D;
}

		}
	}
}