/*---------------------------------------------------------------------



----------------------------------------------------------------------*/
#pragma once
#include <iostream>  
#include <vector>

namespace grid
{
	//Forward declarations
	template<class grid_coord_2d, class cell_2d>
	class BoundedGrid_2d;

	template<class grid_coord_2d, class cell_2d>
	std::ostream& operator<<(std::ostream& os, const BoundedGrid_2d<grid_coord_2d, cell_2d> &grid_2d);


	//
	//
	//
	template<class grid_coord_2d, class cell_2d>
	class BoundedGrid_2d
	{
	public:
		typedef BoundedGrid_2d<grid_coord_2d,cell_2d>                   this_t;

		typedef cell_2d                                                cell_2d;
		typedef grid_coord_2d                                    grid_coord_2d;
		typedef typename grid_coord_2d::component              coord_component;

		typedef std::vector<cell_2d>                            cell_container;
		typedef typename cell_container::size_type                   size_type;
		typedef typename cell_container::iterator                     iterator;
		typedef typename cell_container::const_iterator         const_iterator;


		// construct grid without any cells and a initial resolution of 0,0
		BoundedGrid_2d() : h(1.0f)
		{
		}

		// construct a grid with the given initial resolution
		BoundedGrid_2d( size_type dimX, size_type dimY ) : h(1.0f)
		{
			resize( dimX, dimY );
		}

		// returns a direct reference to the coordinate-specified cell
		iterator getCell( const grid_coord_2d &coordinate )
		{
			// check range
			if( (coordinate.i()>=(coord_component)0)&&(coordinate.i()<(coord_component)m_dimensionX)&&(coordinate.j()>=(coord_component)0)&&(coordinate.j()<(coord_component)m_dimensionY))
				return begin()+hash(coordinate);
			return end();
		}

		// returns a direct reference to the coordinate-specified cell
		iterator getCell( const coord_component &i, const coord_component &j )
		{
			grid_coord_2d coordinate(i,j);
			// check range
			if( (coordinate.i()>=(coord_component)0)&&(coordinate.i()<(coord_component)m_dimensionX)&&(coordinate.j()>=(coord_component)0)&&(coordinate.j()<(coord_component)m_dimensionY))
				return begin()+hash(coordinate);
			return end();
		}

		// returns a direct reference to the coordinate-specified cell
		const_iterator getCell( const grid_coord_2d &coordinate ) const
		{
			// check range
			if( (coordinate.i()>=(coord_component)0)&&(coordinate.i()<(coord_component)m_dimensionX)&&(coordinate.j()>=(coord_component)0)&&(coordinate.j()<(coord_component)m_dimensionY))
				return begin()+hash(coordinate);
			return end();
		}

		// returns number of cells
		size_type size() const
		{
			return m_cells.size();
		}

		// returns width/height of a single cell
		float cellSize() const
		{
			return h;
		}

		// resolution of the grid in x
		size_type dimensionX() const
		{
			return m_dimensionX;
		}

		// resolution of the grid in y
		size_type dimensionY() const
		{
			return m_dimensionY;
		}

		// resizes the grid to the specified dimensions
		void resize( size_type dimX, size_type dimY )
		{
			if( (m_dimensionX == dimX) && (m_dimensionY == dimY) )
				// dont need resize
				return;

			m_dimensionX = dimX;
			m_dimensionY = dimY;

			m_cells.resize( m_dimensionX*m_dimensionY );

			// rebuild cell references
			for( size_type j=0; j<m_dimensionY; ++j )
				for( size_type i=0; i<m_dimensionX; ++i )
				{
					// coordinates
					grid_coord_2d c( (coord_component)i, (coord_component)j ); // center
					cell_2d &cell = *getCell(c);

					// iterate over all neighbours
					for( int t=0; t<8; ++t )
					{
						iterator it = getCell(c.neighbourCoord(t));

						if( it != end() )
							cell.setNeighbour( t, &(*it) );
					}
				}
		}

		// returns iterator (== reference) to the first element
		iterator begin()
		{
			return m_cells.begin();
		}

		// returns iterator (== reference) to the first element
		const_iterator begin() const
		{
			return m_cells.begin();
		}
		
		// returns iterator (== reference) to the first past last element
		iterator end()
		{
			return m_cells.end();
		}

		// returns iterator (== reference) to the first past last element
		const_iterator end() const
		{
			return m_cells.end();
		}

		// returns linear hash value for the specified coordinate
		size_type hash( const grid_coord_2d& key ) const
		{
			return key.j()*m_dimensionX + key.i();
		}

		// returns linear hash value for the specified coordinate
		size_type hash( coord_component i, coord_component j ) const
		{
			return j*m_dimensionX + (size_type)i;
		}

	private:
		cell_container                                      m_cells;
		float                                                     h; // width of a cell
		size_type                                      m_dimensionX;
		size_type                                      m_dimensionY;
	};



	// serialize
	template<class grid_coord_2d, class cell_2d>
	std::ostream& operator<<(std::ostream& os, const BoundedGrid_2d<grid_coord_2d, cell_2d> &grid_2d)
	{
		typedef BoundedGrid_2d<grid_coord_2d, cell_2d>       this_t;
		typedef typename this_t::size_type                size_type;
		
		float h = grid_2d.cellSize();
		size_type dx = grid_2d.dimensionX();
		size_type dy = grid_2d.dimensionX();

		os.write( (char*)&h,sizeof(float));
		os.write( (char*)&dx,sizeof(size_type));
		os.write( (char*)&dy,sizeof(size_type));

		// serialize each cell
		for( this_t::const_iterator it=grid_2d.begin(); it != grid_2d.end(); ++it )
			os<<*it;

		return os;
	}

	// deserialize
	template<class grid_coord_2d, class cell_2d>
	std::istream& operator>>(std::istream& is, BoundedGrid_2d<grid_coord_2d, cell_2d> &grid_2d)
	{
		typedef BoundedGrid_2d<grid_coord_2d, cell_2d>       this_t;
		typedef typename this_t::size_type                size_type;
		
		size_type dx = 0;
		size_type dy = 0;
		float h = 1.0f;

		is.read( (char*)&h,sizeof(float));
		is.read( (char*)&dx,sizeof(size_type));
		is.read( (char*)&dy,sizeof(size_type));

		grid_2d.resize( dx, dy );

		// deserialize each cell
		for( this_t::iterator it=grid_2d.begin(); it != grid_2d.end(); ++it )
			is>>*it;

		return is;
	}
}