/*---------------------------------------------------------------------

Performs FastSweeping

----------------------------------------------------------------------*/
#pragma once


namespace grid
{
	// performs a sweep udate from given neighbours
	// cell: the cell to update
	// val : function object which retrieves the value to be swept from a cell reference
	// hNeighbour/vNeighbour : index of the vertical and horizontal neighbour to sweep from
	// h: cell width must be provide to compute plausible distance values
	// infinite: a value which represents a infinite far distance (used for non existing cells)
	template<class Cell_2d, class Value>
	void sweepUpdate( Cell_2d &cell, Value val, int hNeighbour, int vNeighbour, float h, float infinite=999999999.9f )
	{
		float &cval = val(cell); // value reference of the center cell
		float            newVal;     // new value
		float aDist = infinite; // distance value of first neighbour
		float bDist = infinite; // distance value of second neighbour

		// try to access neighbours
		if( cell.neighbour(hNeighbour) )
			aDist = val(*cell.neighbour(hNeighbour));
		if( cell.neighbour(vNeighbour) )
			bDist = val(*cell.neighbour(vNeighbour));

		// sort neighbour distances in ascending order
		if( fabs(aDist) > fabs(bDist)) std::swap(aDist, bDist);

		
		newVal = aDist + h; // the new value is computed from the smaller neighbour value plus the distance to the current cell (h)

		if( fabs(newVal) > fabs(bDist) )
			newVal = (aDist+bDist+sqrt(2.0f*h*h-(aDist-bDist)*(aDist-bDist)))/2.0f;

		if( newVal < cval )
			cval = newVal;
	}

	// Performs fast sweeping over the range [(i0,j0);(i1,j1)]
	template<class Grid_2d, class CoordComponent, class Value>
	void fast_sweep_2d( Grid_2d &grid, CoordComponent _i0, CoordComponent _j0, CoordComponent _i1, CoordComponent _j1, Value val )
	{
		typedef typename Grid_2d::cell_2d                              cell_2d;
		typedef typename Grid_2d::grid_coord_2d                  grid_coord_2d;

		CoordComponent i0 = _i0;
		CoordComponent j0 = _j0;
		CoordComponent i1 = _i1;
		CoordComponent j1 = _j1;

		if( i0 > i1  )std::swap(i0,i1);
		if( j0 > j1  )std::swap(j0,j1);

		float left = 0.0f;
		float right = 0.0f;
		float bottom = 0.0f;
		float top = 0.0f;

		float h = grid.cellSize();


		// compute remaining(unkown distances) with the fast sweeping algorithm
		// fast sweeping algorithm: iterations
		for(int k=0; k<2; ++k)
		{
			// we do 4 sweeps (one for each quadrant) and update the values

			// Quadrant I   : i 0->1 | j 0->1
			for( long j=j0; j<j1; ++j )
				for( long i=i0; i<i1; ++i )
				{
					// get cell
					Grid::iterator cellIt = grid.getCell(i,j);

					// cell found
					if( cellIt != grid.end() )
					{
						cell_2d &cell = *cellIt;

						sweepUpdate(cell, val, 0, 2, h);
					}
				}

			// Quadrant II   : i 1->0 | j 0->1
			for( long j=j0; j<j1; ++j )
				for( long i=i1; i>=i0; --i )
				{
					// get cell
					Grid::iterator cellIt = grid.getCell(i,j);

					// cell found
					if( cellIt != grid.end() )
					{
						cell_2d &cell = *cellIt;

						sweepUpdate(cell, val, 1, 2, h);
					}
				}

			// Quadrant III   : i 1->0 | j 1->0
			for( long j=j1; j>=j0; --j )
				for( long i=i1; i>=i0; --i )
				{
					// get cell
					Grid::iterator cellIt = grid.getCell(i,j);

					// cell found
					if( cellIt != grid.end() )
					{
						cell_2d &cell = *cellIt;

						sweepUpdate(cell, val, 1, 3, h);
					}
				}

			// Quadrant IV   : i 0->1 | j 1->0
			for( long j=j1; j>=j0; --j )
				for( long i=i0; i<i1; ++i )
				{
					// get cell
					Grid::iterator cellIt = grid.getCell(i,j);

					// cell found
					if( cellIt != grid.end() )
					{
						cell_2d &cell = *cellIt;

						sweepUpdate(cell, val, 0, 3, h);
					}
				}
		}
	}
}