#include "PCAfluid.h"
#include <iostream>
#include <fstream>
#include <sstream>
#include <ctime>
#include <Windows.h>
#include <Commdlg.h>

#define LOGFILE "pca.log"

// helper function to convert from unicode to ansi, copied from msdn lib, needed for load/save dialog boxes
HRESULT UnicodeToAnsi(LPCOLESTR pszW, LPSTR* ppszA)
{
    ULONG cbAnsi, cCharacters;
    DWORD dwError;

    // If input is null then just return the same.
    if (pszW == NULL)
    {
        *ppszA = NULL;
        return NOERROR;
    }

    cCharacters = wcslen(pszW)+1;
    // Determine number of bytes to be allocated for ANSI string. An
    // ANSI string can have at most 2 bytes per character (for Double
    // Byte Character Strings.)
    cbAnsi = cCharacters*2;

    // Use of the OLE allocator is not required because the resultant
    // ANSI  string will never be passed to another COM component. You
    // can use your own allocator.
    *ppszA = (LPSTR) CoTaskMemAlloc(cbAnsi);
    if (NULL == *ppszA)
        return E_OUTOFMEMORY;

    // Convert to ANSI.
    if (0 == WideCharToMultiByte(CP_ACP, 0, pszW, cCharacters, *ppszA,
                  cbAnsi, NULL, NULL))
    {
        dwError = GetLastError();
        CoTaskMemFree(*ppszA);
        *ppszA = NULL;
        return HRESULT_FROM_WIN32(dwError);
    }
    return NOERROR;
}


//helper function: mean squared error MSE = 1/length * ||op1-op2||^2  = 1/length * <op1-op2,op1-op2>. if length of operands not equal, the resulting vector has length of operand w/ smaller one
float mse(std::vector<float> orig, std::vector<float> reconstruction)
{	
	float result = 0.0f;
	float norm = 0.0f;

	//check for length of resulting vector 
	int minlength = (orig.size() < reconstruction.size()) ? orig.size() : reconstruction.size();

	// calculate MSE and euclidian norm of original data
	for(int i=0; i < minlength; i++)
	{
		result += pow(orig[i] - reconstruction[i],2);		
	}
	//return result/norm;
	return result/((float) minlength);    // do we really have to divide by minlength?
}


namespace pca
{	
	PCAfluid::PCAfluid()
	{
		//dummy initialization
		initialize(0, 0, 0, 0, 0);		
	}

	// some overloaded constructors 
	PCAfluid::PCAfluid(int dimX, int dimY, int nrOfBasisStates)
	{
		initialize(dimX, dimY, 1, nrOfBasisStates, false);
	}

	PCAfluid::PCAfluid(int dimX, int dimY, int nrOfBasisStates, bool UseTransposed)
	{
		initialize(dimX, dimY, 1, nrOfBasisStates, UseTransposed);
	}

	PCAfluid::PCAfluid(int dimX, int dimY, int dimZ, int nrOfBasisStates)
	{
		initialize(dimX, dimY, dimZ, nrOfBasisStates, false);
	}

	PCAfluid::PCAfluid(int dimX, int dimY, int dimZ, int nrOfBasisStates, bool UseTransposed)
	{
		initialize(dimX, dimY, dimZ, nrOfBasisStates, UseTransposed);
	}

	PCAfluid::~PCAfluid()
	{
	}


	// overloaded initialize routines to set up different values like data extents or file import options
	void PCAfluid::initialize(int dimX, int dimY, int nrOfBasisStates)
	{
		initialize(dimX, dimY, 1, nrOfBasisStates, false);
	}

	void PCAfluid::initialize(int dimX, int dimY, int nrOfBasisStates, bool UseTransposed)
	{
		initialize(dimX, dimY, 1, nrOfBasisStates, UseTransposed);
	}

	void PCAfluid::initialize(int dimX, int dimY, int dimZ, int nrOfBasisStates)
	{
		initialize(dimX, dimY, dimZ, nrOfBasisStates, false);
	}

	void PCAfluid::initialize(int dimX, int dimY, int dimZ, int nrOfBasisStates, bool UseTransposed)
	{
		m_iDimX = dimX;
		m_iDimY = dimY;
		m_iDimZ = dimZ;
		m_iDataSize = dimX * dimY * dimZ;

		m_bUseTransposed = UseTransposed;
		m_bRecalcReconstructionError = true;	
		
		m_iNrOfDataSets = 0;
		m_iNrOfBasisStates = nrOfBasisStates;
		m_fvPrincCompMat = new cgv::math::mat<float>();		
	}


	// appends another data set to current data set matrix
	void PCAfluid::appendDataSet(std::vector<float> data)
	{		
		// check for correct dimensions
		if(data.size() != m_iDataSize)
		{
			std::cout<<"ups...could not append data set, wrong dimensions"<<std::endl;
			return;
		}				
		//copy data into temp vector
		std::vector<float> newData(data);
		
		//append new data vector to existing matrix
		m_fvDataMat.push_back(newData);
		m_iNrOfDataSets++;
		std::cout<<"\rNumber of data sets: "<<m_iNrOfDataSets;		
	}

	//main algorithm
	void PCAfluid::executePCA()
	{	
		// start a time measurement for statistics
		std::cout<<std::endl<<"Calculating, please wait..."<<std::endl;
		std::clock_t start = std::clock();

		// copy datasets into one big data set matrix
		cgv::math::mat<float> *data = new cgv::math::mat<float>();
		data->resize(m_iDataSize,m_iNrOfDataSets);

		//iterate over all datasets
		for (int i = 0; i < m_iNrOfDataSets; i++)
		{
			//iterate over elements of single data set
			for (int j = 0; j < m_iDataSize; j++)
			{
				//pay attention of indices (data sets are given as column vectors)
				(*data)(j,i) = m_fvDataMat[i][j];
			}
		}

		// if necessary transpose the entire data set matrix		
		if (m_bUseTransposed)
		{
			(*data) = cgv::math::transpose(*data); 
		}

		//  calculate the means over rows of data set matrix
		m_fvMeanValues.resize(data->nrows(), 0.0f);
		cgv::math::vec<float> idVec = cgv::math::ones<float>(data->ncols(),1).col(0); //help vector of size data->ncols(), all entries set to one
		for (unsigned i = 0; i < data->nrows(); i++)
		{
			m_fvMeanValues[i] = cgv::math::dot(data->row(i),idVec)/data->ncols(); // sum of all row elements equals dot product of particular row vector with help vector
		}

		std::cout<<"Mean Values calculated."<<std::endl;

		// subtract row mean from every entry in corresponding row
		for (int i = 0; i < data->nrows(); i++)
		{
			for (int j = 0; j < data->ncols(); j++)
			{
				(*data)(i,j) = (*data)(i,j) - m_fvMeanValues[i];
			}
		}
		
		// data matrix is centered, let's do the SVD 
		std::cout<<"Matrix prepared. Waiting for SVD decomposition result..."<<std::endl;		
		
		
		// svd of matrix Y = 1/sqrt(nr of cols - 1) * Data^T = U*S*V^T
		// V then contains the principal components of covariance matrix Data*Data^T (as column vectors)		
		cgv::math::mat<float> *data2 = new cgv::math::mat<float>();
		(*data2) = 1.0f/(sqrtf((float)data->ncols() - 1.0f))*cgv::math::transpose(*data); 
		
		cgv::math::diag_mat<float> *s = new cgv::math::diag_mat<float>();
		s->resize(data->nrows());

		m_fvPrincCompMat->resize(data->nrows(),data->nrows());
	
		if(cgv::math::svd(*data2,*data2,*s,*m_fvPrincCompMat,true,20))
		{
			std::cout<<"principal components successfull computed"<<std::endl;			
			//compute sum of diagonal entries of s, we need it to normalize the eigenvalues afterwards
			float diagSum = 0.0f;
			for (unsigned i=0; i<s->ncols(); i++)
			{
				diagSum += (*s)(i);
			}			

			// if data was not transposed preserve only the first m_iNrOfDataSets Principal Components
			if (!m_bUseTransposed)
			{
				cgv::math::mat<float> temp;
				temp.resize(m_iDataSize,m_iNrOfDataSets);
				m_fvPrincCompMat->copy(0,0,m_iDataSize,m_iNrOfDataSets,temp);
				m_fvPrincCompMat->resize(m_iDataSize,m_iNrOfDataSets);
				temp.copy(0,0,temp.nrows(),temp.ncols(),*m_fvPrincCompMat);
			}

			// save the eigen values and print  out
			m_fvEigenValues.resize(m_fvPrincCompMat->ncols());
			for (int i=0; i < m_fvPrincCompMat->ncols(); i++)
			{
				m_fvEigenValues[i] = (*s)(i)/diagSum; // we save the normalized eigenvalues
				std::cout<<i+1<<". principal component: weight = "<<(float)((int)((*s)(i)/diagSum*10000))/100<<"%, length = "<<cgv::math::length(m_fvPrincCompMat->col(i))<<std::endl;
			}
		}
		else
		{
			std::cout<<"something has gone wrong, what a pity..."<<std::endl;
		}
		// compute compressed data sets (respectively the feature matrix) 
		computeCompressedDataSets(0,m_iNrOfBasisStates);

		// elapsed calculation time
		std::clock_t end = std::clock();
		double elapsedTime = (end - start)/CLOCKS_PER_SEC;
		std::cout<<"...finished in "<<elapsedTime<< "s"<<std::endl;

		//write computation time into the log file
		std::ofstream file(LOGFILE, std::ios::out | std::ios::app );
		if(file.is_open())
		{
			file<<"Nr of data sets:"<<m_iNrOfDataSets<<"    computation time:"<<elapsedTime<<"\n";
			file.close();
		}
		

	}


	const std::vector<float> PCAfluid::getDataSet(int nr) const
	{
		if (m_fvDataMat.size() == 0)  // return dummy data vector if no sets are appended yet
		{
			std::vector<float> temp;
			temp.push_back(0.0f);
			return temp;
		}

		return m_fvDataMat[nr];
	}

	void PCAfluid::computeCompressedDataSets(int lower, int upper)
	{		
		cgv::math::mat<float> origData;
		int nrRows, nrCols;
		
		nrRows = m_iDataSize;
		nrCols = m_iNrOfDataSets;	
		
		if (m_bUseTransposed)
		{
			nrRows = m_iNrOfDataSets;
			nrCols = m_iDataSize;			
		}		

		origData.resize(nrRows, nrCols);	

		//iterate over all datasets
		for (int i = 0; i < origData.nrows(); i++)
		{
			//iterate over elements of single data set
			for (int j = 0; j < origData.ncols(); j++)
			{
				//pay attention of indices
				if (m_bUseTransposed)
					origData(i,j) = m_fvDataMat[i][j] - m_fvMeanValues[i];
				else 
					origData(i,j) = m_fvDataMat[j][i] - m_fvMeanValues[i];
			}
		}

		// we resize PC matrix, copy in the appropriate PCs and compute feature matrix with matrix multiplication
		m_fvFirstPrincCompMat.resize(nrRows,upper-lower);
		m_fvPrincCompMat->copy(0,lower,nrRows,upper-lower,m_fvFirstPrincCompMat);
		m_fvCompressedDataMat = cgv::math::transpose(m_fvFirstPrincCompMat)*origData;


		// if we are using the transposed variant of the PCA save all reconstructed data vectors explicitly to ensure performance
		if (m_bUseTransposed)
		{
			m_transposedReconstruction.clear();
			cgv::math::mat<float> temp;
			temp = m_fvFirstPrincCompMat * m_fvCompressedDataMat;	
	
			for(int i = 0; i < temp.nrows(); i++)
			{
				std::vector<float> dataset;
				for(int j = 0; j < temp.ncols(); j++)
				{
					dataset.push_back(temp(i,j)+m_fvMeanValues[i]);
				}
				m_transposedReconstruction.push_back(dataset);
			}
		}
	}

	const std::vector<float> PCAfluid::getReconstructedDataSet(int nr) const
	{
		// if we are in "transpose mode" we can directly return the reconstructed data
		if (m_bUseTransposed)
			return m_transposedReconstruction[nr];		
		// else we compute it by matrix multiplication of corresponding features with PC matrix
		cgv::math::vec<float> tempVec;
		std::vector<float> data;
		tempVec = m_fvFirstPrincCompMat * m_fvCompressedDataMat.col(nr);
		for (unsigned i=0; i<tempVec.size();i++)
		{
			data.push_back(tempVec(i) + m_fvMeanValues[i]);
		}

		return data;
	}


	void PCAfluid::setNrOfBasisStates(int nr)
	{
		setBasisStatesInterval(0, nr);
	}

	void PCAfluid::setBasisStatesInterval(int lower, int upper)
	{				
				computeCompressedDataSets(lower, upper);
				m_iNrOfBasisStates = upper - lower;
				std::cout<<"Number of basis states set to: "<<lower<<", "<<upper<<std::endl;
				m_bRecalcReconstructionError = true;		
	}

	const int PCAfluid::getNrOfBasisStates() const
	{
		return m_iNrOfBasisStates;
	}

	const int PCAfluid::getNrOfDataSets() const
	{
		return m_iNrOfDataSets;
	}

	std::vector<float> PCAfluid::getPrincipalComponent(int nr) const
	{
		std::vector<float> tempComp;
		// if we have done the PCA with transposed data we return a dummy vector
		if (m_bUseTransposed)
		{
			for (int i=0; i<tempComp.size();i++)
			{
				tempComp.push_back(0.0f);
			}
			return tempComp;
		}		

		// else we build a new (column) vector from PC matrix
		tempComp.resize(m_fvPrincCompMat->nrows());
		if(nr >= 0 && nr < m_fvPrincCompMat->ncols())
		{			
			for (unsigned i = 0; i < m_fvPrincCompMat->nrows(); i++)
			{
				tempComp[i] = (*m_fvPrincCompMat)(i,nr);
			}	
		}
		else 
		{
			std::cout<<"Error! Number of Principal Component must not be smaller than zero either larger than "<<getNrOfPrincipalComponents()-1<<"!"<<std::endl;			
		}		
		return tempComp;		
	}

	const int PCAfluid::getNrOfPrincipalComponents() const
	{
		return m_fvPrincCompMat->ncols();		
	}

	// returns all the features corresponding to one particular PC
	const std::vector<float> PCAfluid::getFeatureVector(int NrOfPrincipalComponent) const
	{
		std::vector<float> result;
		if(NrOfPrincipalComponent >= 0 && NrOfPrincipalComponent < m_fvCompressedDataMat.nrows())
		{
			for(int i=0; i<m_fvCompressedDataMat.ncols(); i++)
				result.push_back(m_fvCompressedDataMat(NrOfPrincipalComponent,i));	
		}
		else
			std::cout<<"Error! Number of Principal Component must not be smaller than zero either larger than "<<getNrOfBasisStates()<<"!"<<std::endl;

		return result;
	}

	// returns all the features corresponding to one particular data set
	const std::vector<float> PCAfluid::getFeaturesOfDataSet(int NrOfDataSet) const
	{
		std::vector<float> result;
		for(int i=0; i < m_fvCompressedDataMat.nrows(); i++)
		{
			result.push_back(m_fvCompressedDataMat(i,NrOfDataSet));
		}
		return result;
	}

	// saves member variables to file
	void PCAfluid::save(const char *filename) const
	{		
		//opens new file to write into (overwrites any files with same name)
		std::ofstream file(filename, std::ios::out | std::ios::binary | std::ios::trunc );
		if(file.is_open())
		{
			std::cout<<"Saving, please wait... ";
			//save primitive types
			file.write((char *) &m_iDimX, sizeof(int));	
			file.write((char *) &m_iDimY, sizeof(int));
			file.write((char *) &m_iDimZ, sizeof(int));
			file.write((char *) &m_iDataSize, sizeof(int));
			file.write((char *) &m_iNrOfBasisStates, sizeof(int));
			file.write((char *) &m_iNrOfDataSets, sizeof(int));			
			file.write((char *) &m_bUseTransposed, sizeof(bool));
		
			//save all data sets
			int w = m_fvDataMat.size();
			int h = m_fvDataMat[0].size();

			file.write((char *) &w, sizeof(int));
			file.write((char *) &h, sizeof(int));	

			for (int i=0; i < w; i++)
			{
				for(int j=0; j < h; j++)
				{
					float temp = m_fvDataMat[i][j];
					file.write((char *) &temp, sizeof(float));
				}				
			}
				
			//save all row mean values
			int size = m_fvMeanValues.size();
			file.write((char *) &size, sizeof(int));

			for (int i=0; i < size; i++)
			{
				float temp = m_fvMeanValues[i];
				file.write((char *) &temp, sizeof(float));
			}
			
			//save the matrix containing the principal components
			int rows = m_fvPrincCompMat->nrows();
			int cols = m_fvPrincCompMat->ncols();

			file.write((char *) &rows, sizeof(int));
			file.write((char *) &cols, sizeof(int));

			for (int i=0; i < rows; i++)
			{
				for (int j=0; j < cols; j++)
				{
					float temp = (*m_fvPrincCompMat)(i,j);
					file.write((char *) &temp, sizeof(float));
				}
			}

			//save eigen values
			int nr_eigenvalues = m_fvEigenValues.size();
			file.write((char *) &nr_eigenvalues, sizeof(int));

			for(int i=0; i < nr_eigenvalues; i++)
			{
				float temp = m_fvEigenValues[i];
				file.write((char *) &temp, sizeof(float));
			}		

			std::cout<<"finished!"<<std::endl;
		}
		else
		{
			std::cout<<"Error! Could not save to file!"<<std::endl;
		}
		file.close();
	}

	// save dialog box
	void PCAfluid::save() const
	{
		// copied from msdn lib
		OPENFILENAME ofn;       // common dialog box structure
		char szFile[512];       // buffer for file name
		
		// Initialize OPENFILENAME
		ZeroMemory(&ofn, sizeof(ofn));
		ofn.lStructSize = sizeof(ofn);
		ofn.lpstrFile = (LPWSTR) szFile;
		// Set lpstrFile[0] to '\0' so that GetOpenFileName does not 
		// use the contents of szFile to initialize itself.
		ofn.lpstrFile[0] = '\0';
		ofn.nMaxFile = sizeof(szFile);
		ofn.lpstrFilter = L"PCA files\0*.pca\0";
		ofn.nFilterIndex = 1;
		ofn.lpstrFileTitle = NULL;
		ofn.nMaxFileTitle = 0;
		ofn.lpstrInitialDir = NULL;
		ofn.Flags = OFN_PATHMUSTEXIST;

		// Display the Open dialog box. 
		GetSaveFileName(&ofn);

		LPSTR temp;
		UnicodeToAnsi(ofn.lpstrFile, &temp);
		std::string fileName = temp;
		fileName += ".pca";
		save(fileName.c_str());
	}

	// load a state from file
	void PCAfluid::load(const char *filename) 
	{	
	    //opens the file and overwrites the member variables
		std::ifstream file(filename, std::ios::in | std::ios::binary);
		if(file.is_open())
		{
			std::cout<<"Loading file, please wait... ";
			file.seekg (0, std::ios::beg);
			
			//restore primitive types
			file.read((char *) &m_iDimX, sizeof(int));	
			file.read((char *) &m_iDimY, sizeof(int));
			file.read((char *) &m_iDimZ, sizeof(int));
			file.read((char *) &m_iDataSize, sizeof(int));
			file.read((char *) &m_iNrOfBasisStates, sizeof(int));
			file.read((char *) &m_iNrOfDataSets, sizeof(int));
			file.read((char *) &m_bUseTransposed, sizeof(bool));
						
			//restore data sets
			int w,h;
			file.read((char *) &w, sizeof(int));
			file.read((char *) &h, sizeof(int));			

			m_fvDataMat.clear();
			//m_fvDataMat.resize(w);
			
			for (int i=0; i < w; i++)
			{
				std::vector<float> tempVec;
				tempVec.resize(h);
				m_fvDataMat.push_back(tempVec);				

				for(int j=0; j < h; j++)
				{
					float temp;
					file.read((char *) &temp, sizeof(float));
					m_fvDataMat[i][j]=temp;
				}					
			}
				
			//restore row mean values
			int size;
			file.read((char *) &size, sizeof(int));

			m_fvMeanValues.clear();
			m_fvMeanValues.resize(size);

			for (int i=0; i < size; i++)
			{
				float temp;
				file.read((char *) &temp, sizeof(float));
				m_fvMeanValues[i] = temp;
				//std::cout<<i<<". mean value: "<<temp<<std::endl;
			}
			
			//restore the matrix containing the principal components
			int rows,cols;
			file.read((char *) &rows, sizeof(int));
			file.read((char *) &cols, sizeof(int));

			m_fvPrincCompMat->resize(rows,cols); 				
			
			for (int i=0; i < rows; i++)
			{
				for (int j=0; j < cols; j++)
				{
					float temp;
					file.read((char *) &temp, sizeof(float));
					m_fvPrincCompMat->operator ()(i,j) = temp;					
				}
			}

			//restore eigen values
			int nr_eigenvalues;
			file.read((char *) &nr_eigenvalues, sizeof(int));

			m_fvEigenValues.clear();
			m_fvEigenValues.resize(nr_eigenvalues);
			for(int i=0; i < nr_eigenvalues; i++)
			{
				float temp;
				file.read((char *) &temp, sizeof(float));
				m_fvEigenValues[i] = temp;
				//std::cout<<i<<". eigen value: "<<temp<<std::endl;
			}					
			
			std::cout<<"finished!"<<std::endl;			
		}
		else
		{
			std::cout<<"Error! Could not load from file!"<<std::endl;
		}
		file.close();
		
		computeCompressedDataSets(0, m_iNrOfBasisStates);
	}

	// load dialog box
	void PCAfluid::load()
	{
		// copied from msdn lib
		OPENFILENAME ofn;       // common dialog box structure
		char szFile[512];       // buffer for file name
		
		// Initialize OPENFILENAME
		ZeroMemory(&ofn, sizeof(ofn));
		ofn.lStructSize = sizeof(ofn);
		ofn.lpstrFile = (LPWSTR) szFile;
		// Set lpstrFile[0] to '\0' so that GetOpenFileName does not 
		// use the contents of szFile to initialize itself.
		ofn.lpstrFile[0] = '\0';
		ofn.nMaxFile = sizeof(szFile);
		ofn.lpstrFilter = L"PCA files\0*.pca\0";
		ofn.nFilterIndex = 1;
		ofn.lpstrFileTitle = NULL;
		ofn.nMaxFileTitle = 0;
		ofn.lpstrInitialDir = NULL;
		ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST;

		// Display the Open dialog box. 
		GetOpenFileName(&ofn);

		LPSTR temp;
		UnicodeToAnsi(ofn.lpstrFile, &temp);
		char *fileName = temp;

		load(fileName);
	}


	const int PCAfluid::getDimX() const
	{
		return m_iDimX;
	}

	const int PCAfluid::getDimY() const
	{
		return m_iDimY;
	}

	const int PCAfluid::getDimZ() const
	{
		return m_iDimZ;
	}

	const int PCAfluid::getDataSetSize() const
	{
		return m_iDimX * m_iDimY * m_iDimZ;
	}

	const float PCAfluid::getCompressionRatio() const
	{
		float orig, comp;
		orig = (float) getDataSetSize()*getNrOfDataSets();
		comp = (float) getNrOfBasisStates()*(getDataSetSize()+getNrOfDataSets());

		return orig/comp;		
	}
	

	const std::vector<float> PCAfluid::getEigenValues() const
	{
		return m_fvEigenValues;
	}


	const float PCAfluid::getReconstructionError(int NrOfSet)
	{
		return mse(getDataSet(NrOfSet),getReconstructedDataSet(NrOfSet));
	}


	const float PCAfluid::getMeanReconstructionError()
	{
		if (m_bRecalcReconstructionError)
		{
			calculateReconstructionErrors();
		}	
		return m_fMeanReconstructionError;		
	}

	const float PCAfluid::getMinReconstructionError()
	{
		if (m_bRecalcReconstructionError)
		{
			calculateReconstructionErrors();
		}
		return m_fMinReconstructionError;
	}


	const float PCAfluid::getMaxReconstructionError()
	{
		if (m_bRecalcReconstructionError)
		{
			calculateReconstructionErrors();
		}
		return m_fMaxReconstructionError;
	}

	// here we calculate all error terms at one time
	void PCAfluid::calculateReconstructionErrors()
	{
		float sum=0.0f, max=0.0f, min = mse(getDataSet(0),getReconstructedDataSet(0));

		for(int i=0; i<m_iNrOfDataSets; i++)
		{
			float temp = mse(getDataSet(i),getReconstructedDataSet(i));
			min = (temp<min) ? temp : min;
			max = (temp>max) ? temp : max;
			sum += temp;			
		}

		m_fMinReconstructionError = min;
		m_fMaxReconstructionError = max;
		m_fMeanReconstructionError = sum/((float) m_iNrOfDataSets);
		m_bRecalcReconstructionError = false;
	}

	//Linear interpolation between feature vectors
	const std::vector<float> PCAfluid::getInterpolatedReconstructedDataSet(int first, int second, float t) const
	{
		if (m_bUseTransposed)
			return m_transposedReconstruction[first]; //don't interpolate in case of pca with transposed data		
		
		cgv::math::vec<float> tempVec;
		std::vector<float> data;
		tempVec = m_fvFirstPrincCompMat * ( (1-t) * m_fvCompressedDataMat.col(first) + t * m_fvCompressedDataMat.col(second) ); //<- here we interpolate linear
		for (unsigned i=0; i<tempVec.size();i++)
		{
			data.push_back(tempVec(i) + m_fvMeanValues[i]);
		}
		return data;
	}

	//Linear interpolation between data vectors
	const std::vector<float> PCAfluid::getInterpolatedDataSet(int first, int second, float t) const
	{
		std::vector<float> temp;
		if (m_fvDataMat.size() == 0)  // return dummy data vector if no data sets are appended yet
		{
			temp.push_back(0.0f);
			return temp;
		}
		for(int i = 0; i < m_iDataSize; i++)
		{
			temp.push_back( (1-t) * m_fvDataMat[first][i] + t * m_fvDataMat[second][i] ); //<- here we interpolate linear
		}			
		return temp;
	}

	// normalize data so that the maximum magnitude of scalar values equals 1 
	void PCAfluid::normalizeDataSets()
	{
		//check out max. magnitude
		float max = 0.0f;

		for(int i=0; i < m_iNrOfDataSets; i++)
		{
			for(int j=0; j < m_iDataSize; j++)
			{
				max = (fabsf(m_fvDataMat[i][j]) > max) ? fabsf(m_fvDataMat[i][j]) : max;
			}			
		}		
		
		// divide every data vector entry by max
		for(int i=0; i < m_iNrOfDataSets; i++)
		{
			for(int j=0; j < m_iDataSize; j++)
			{
				m_fvDataMat[i][j] /= max;
			}			
		}
	}



}





