/*-------------------------------------------------------------------

PCAfluid Class provides all functionalities and data structures 
to execute Principal Component Analysis and administrate data sets.

--------------------------------------------------------------------*/

#pragma once
#include <vector>
#include <cgv/math/mat.h>
#include <cgv/math/svd.h>
#include <string>




namespace pca
{
	class PCAfluid
	{
		public:

			//dummy constructor
			PCAfluid();
			
			// constructors for 2D data sets
			PCAfluid(int dimX, int dimY, int nrOfBasisStates);
			PCAfluid(int dimX, int dimY, int nrOfBasisStates, bool UseTransposed);
			
			// constructors for 3D data sets
			PCAfluid(int dimX, int dimY, int dimZ, int nrOfBasisStates);
			PCAfluid(int dimX, int dimY, int dimZ, int nrOfBasisStates, bool UseTransposed);

			~PCAfluid();

		
			int m_iNrOfDataSets;   //how many datasets are used to build up covariance matrix
			int m_iNrOfBasisStates; // how many pairs of eigenvalues/eigenvectors are used after decomposition of covariance matrix
			int m_iDimX, m_iDimY, m_iDimZ; //dimensions of input datasets
			long int m_iDataSize; // size of data vector (nr of elements)
			
			std::vector<std::vector<float>> m_fvDataMat; // matrix containing data sets as column vectors
			cgv::math::mat<float> m_fvCompressedDataMat; // matrix containing the compressed data sets (feature vectors)

			std::vector<float> m_fvMeanValues; // save the mean values for corresponding dataset scalars
			std::vector<float> m_fvEigenValues; // vector containing the eigen values of princ. comp.

			cgv::math::mat<float> *m_fvPrincCompMat; // pointer to matrix storing the principal components as columns
			cgv::math::mat<float> m_fvFirstPrincCompMat; // storing the first <nrOfBasisStates> principal components as columns

			std::vector<std::vector<float>> m_transposedReconstruction;	// if we do a PCA of a transposed data set we explicitly hold
																		// the reconstructed data separately to avoid heavy calculation

			bool m_bUseTransposed; // PCA can also be done with transposed data (needs less memory in case of large data extents)
			bool m_bRecalcReconstructionError; // decides if we have to do a recalculation of all errors

			float m_fMeanReconstructionError;
			float m_fMinReconstructionError;
			float m_fMaxReconstructionError;

			void initialize(int dimX, int dimY, int nrOfBasisStates); // sets up the 2D data set dimensions and initializes vectors
			void initialize(int dimX, int dimY, int nrOfBasisStates, bool UseTransposed);
			void initialize(int dimX, int dimY, int dimZ, int nrOfBasisStates); // sets up the 3D data set dimensions and initializes vectors
			void initialize(int dimX, int dimY, int dimZ, int nrOfBasisStates, bool UseTransposed);

			bool useTransposing(){return m_bUseTransposed;} // do we have transposed the data before executing the PCA algorithm?
					
			void appendDataSet(std::vector<float> data); // appends a data set to the current data set stack
			
			void executePCA(); // main funtion that executes the algorithm

			void computeCompressedDataSets(int lower, int upper); //calculates the data to compress with (maybe reduced) number of basis states

			void setNrOfBasisStates(int nr); // sets the number of the first "nr" PCs (0..nr-1)
			void setBasisStatesInterval(int lower, int upper); // sets an interval of basis states [lower, upper]
			const int getNrOfBasisStates() const; // returns current number of used basis states

			std::vector<float> getPrincipalComponent(int nr) const; // returns the PC with number "nr"
			const int getNrOfPrincipalComponents() const; // returns the count of PCs		

			const std::vector<float> getDataSet(int nr) const; // returns particular data vector
			const std::vector<float> getReconstructedDataSet(int nr) const; // returns particular data vector after reconstruction

			void save(const char* filename) const; //save all members to file
			void save() const; //save state using dialog box
			void load(const char* filename); //restore all members from file
			void load(); // loads state using dialog box

			const int getNrOfDataSets() const; // returns the number of data sets which are currently used

			const int getDimX() const; // returns data extent in x-direction

			const int getDimY() const; // returns data extent in y-direction

			const int getDimZ() const; // returns data extent in z-direction

			const std::vector<float> getFeatureVector(int NrOfPrincipalComponent) const; // returns the vector belonging to particular PC for all data sets

			const std::vector<float> getFeaturesOfDataSet(int NrOfDataSet) const; // returns the vector belonging to particular data set for all PCs

			const int getDataSetSize() const; // returns size of data vector

			const float getCompressionRatio() const; // return compression ratio

			const std::vector<float> getEigenValues() const; // returns ordered set of eigenvalues

			const float getReconstructionError(int NrOfSet); // returns reconstruction error (MSE) for one data vector

			const float getMeanReconstructionError(); // returns mean MSE for all data vectors

			const float getMinReconstructionError(); // returns minimum MSE for all data vectors

			const float getMaxReconstructionError(); // returns maximum MSE for all data vectors

			void calculateReconstructionErrors(); // helper function to calculate all error terms at one time

			std::vector<float> getMeanValues() {return m_fvMeanValues;} // returns the mean values of data vectors

			void normalizeDataSets(); // normalize to interval [-1,1] where 0 describes fluid surface

			const std::vector<float> getInterpolatedReconstructedDataSet(int first, int second, float t) const; // returns the reconstruction of an interpolated data vector

			const std::vector<float> getInterpolatedDataSet(int first, int second, float t) const; // returns an interpolated data vector
	};
}