/*------------------------------------------------------------------------

Implementation of Render Manager that adapts the render algorithm 
to the PCA class.

-------------------------------------------------------------------------*/
#include "RenderManager.h"
#include "math/math.h"
#include <sstream>

#pragma message("     _Adding library: opengl32.lib" ) 
#pragma comment( lib, "opengl32.lib" )

RenderManager::RenderManager(int wwidth, int wheight)
{
	//initialize miscellaneous properties
	viewManipulator.setCamera( &camera );
	WindowWidth = wwidth;
	WindowHeight = wheight;
	MCDisplayListID1 = 0;
	MCDisplayListID2 = 0;
	setMCEnvironment();
	DirtyFlag = true;

	PropertiesAmbient[0] = 0.50f;
	PropertiesAmbient[1] = 0.50f;
	PropertiesAmbient[2] = 0.50f;
	PropertiesAmbient[3] = 1.00f;
	
	PropertiesDiffuse[0] = 0.75f;
	PropertiesDiffuse[1] = 0.75f;
	PropertiesDiffuse[2] = 0.75f;
	PropertiesDiffuse[3] = 1.00f;

	PropertiesSpecular[0] = 1.00f;
	PropertiesSpecular[1] = 1.00f;
	PropertiesSpecular[2] = 1.00f;
	PropertiesSpecular[3] = 1.00f;
	
	Color[0] = 0.27f;
	Color[1] = 0.39f;
	Color[2] = 0.83f;
	Color[3] = 1.00f;	
}

RenderManager::~RenderManager()
{
}


void RenderManager::setMCEnvironment() //setting up colors and materials for Marching Cubes visualization
{	
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
	glPolygonMode(GL_FRONT, GL_FILL);
	glClearColor(1.0f,1.0f,1.0f,1.0f);
			  
	glLightfv( GL_LIGHT0, GL_AMBIENT,  PropertiesAmbient); 
	glLightfv( GL_LIGHT0, GL_DIFFUSE,  PropertiesDiffuse); 
	glLightfv( GL_LIGHT0, GL_SPECULAR, PropertiesSpecular); 
	glLightModelf(GL_LIGHT_MODEL_TWO_SIDE, 1.0); 
	glEnable( GL_LIGHT0 ); 
	
	glMaterialfv(GL_BACK,  GL_AMBIENT,   Color); 
	glMaterialfv(GL_BACK,  GL_DIFFUSE,   Color); 
	glMaterialfv(GL_BACK, GL_SPECULAR,  PropertiesSpecular);
	glMaterialf( GL_BACK, GL_SHININESS, 25.0); 
	
	glMaterialfv(GL_FRONT, GL_AMBIENT,   Color); 
	glMaterialfv(GL_FRONT, GL_DIFFUSE,   Color); 
	glMaterialfv(GL_FRONT, GL_SPECULAR,  PropertiesSpecular);
	glMaterialf( GL_FRONT, GL_SHININESS, 25.0);
}

void RenderManager::regenerateTextureIDs() // delete old textures, free up memory and generate new ones
{
glDeleteTextures(1, &reconstructedDataTexID);
glDeleteTextures(1, &originalDataTexID);

glGenTextures(1, &reconstructedDataTexID);
glGenTextures(1, &originalDataTexID);
}

void RenderManager::setDirtyFlag() // helper function: used to force a redraw (e.g. after a parameter change)
{
	DirtyFlag = true;
};

void RenderManager::setWindowSize(int width, int height) //helper function to set window size
{
	WindowWidth = width;
	WindowHeight = height;
}

// texture related functions which build a new texture from the given data (with or without clamping)
void RenderManager::buildTextureFromDataVector(int width, int height, const std::vector<float> &dataVector1, const std::vector<float> &dataVector2, bool clamp)
{
	RenderManager::buildTextureFromDataVector(width, height, originalDataTexture, originalDataTexID, dataVector1, clamp);
	RenderManager::buildTextureFromDataVector(width, height, reconstructedDataTexture, reconstructedDataTexID, dataVector2, clamp);
}
void RenderManager::buildTextureFromDataVector(int width, int height, util::Image &texture, unsigned int &texID, const std::vector<float> &dataVector, bool clamp)
{	
	texture.setResolution(width,height);

	// find min/max signed distance (for visualization w/o clamping)
	float minDistance = 0.0f;
	float maxDistance = 0.0f;	

	for( int i=0; i < width*height; i++ )
	{
		float temp = dataVector[i];
	
		if( temp > maxDistance )
			maxDistance = temp;
		if( temp < minDistance )
			minDistance = temp;
	}
		
	// iterate over cells
	for( int v=0; v < height; v++ )
	{
		for( int u=0; u < width; u++ )
		{						
			float val;

			//clamp to [-1..1]
			if (clamp)
			{
				float value = dataVector[v*width + u];
				if(dataVector[v*width + u] < -1.0)
					value = -1.0;
				if(dataVector[v*width + u] > 1.0)
					value = 1.0;
				val = math::mapValueToRange( -1.0, 1.0, 0.0f, 1.0f, value );
			}
			
			else
			{
				//without clamping we normalize over the interval [minDistance,maxDistance]  
				val = math::mapValueToRange( minDistance, maxDistance, 0.0f, 1.0f, dataVector[v*width + u] ); 
			}
			
			math::Color pixelColor( val, val, val );
			texture.setPixel( u, v, math::setColor(pixelColor) ); //finally set the pixelColor as grey scale
		}
	}

	// upload texture
	glGenTextures(1,&texID);
	glBindTexture( GL_TEXTURE_2D, texID);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE );
	gluBuild2DMipmaps(GL_TEXTURE_2D, 1, texture.getWidth(), texture.getHeight(), GL_RGBA, GL_UNSIGNED_BYTE, (unsigned char*)texture.getImageBuffer());	
}

// update view in case of dirty frame 
void RenderManager::rebuildTextures(int DataSetNr, bool showRecalculatedData, int princCompNr, bool useClamping) 
{
	regenerateTextureIDs();
	int DimX = PCAComponent->getDimX();
	int DimY = PCAComponent->getDimY();

	buildTextureFromDataVector(DimX,DimY, originalDataTexture, originalDataTexID, PCAComponent->getDataSet(DataSetNr), useClamping);
		
	if(showRecalculatedData)
		buildTextureFromDataVector(DimX, DimY, reconstructedDataTexture, reconstructedDataTexID, PCAComponent->getReconstructedDataSet(DataSetNr),useClamping);
		
	else 
		buildTextureFromDataVector(DimX, DimY, reconstructedDataTexture, reconstructedDataTexID, PCAComponent->getPrincipalComponent(princCompNr),false); //don't clamp values to -1..1 when showing PCs
}

// helper function to render one line of text with GLUT Bitmap font renderer with specified position, font and color
void RenderManager::renderTextLine(float XPos, float YPos, void *font, std::string text, float r, float g, float b)
{
	const char *c = text.c_str();
	glColor3f(r,g,b);
	glRasterPos2f(XPos, YPos);
	while (*c != '\0')
	{
		glutBitmapCharacter(font, *c);
		c++;
	}
}

// renders one line of text with standard font and color
void RenderManager::renderTextLine(float XPos, float YPos, std::string text)
{
	renderTextLine(XPos, YPos, GLUT_BITMAP_HELVETICA_12, text, 0.0f, 0.0f, 0.0f);
}


// register the pca object
void RenderManager::setPCAComponent(pca::PCAfluid *component)
{
	PCAComponent = component;
};

// returns the current pca object
pca::PCAfluid* RenderManager::getPCAComponent()
{	
	return PCAComponent;
}

// render the data with textures
void RenderManager::renderTextures()
{
	glDisable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);
	
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	glOrtho(0,2,0,1,-1,1);
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	glColor3f(0.0f,0.0f,0.0f);
	glBegin(GL_LINES);
	glVertex2d(1,1);
	glVertex2d(1,0);
	glEnd();

	glBindTexture(GL_TEXTURE_2D, originalDataTexID);
	glBegin(GL_QUADS);
	glTexCoord2f(0.0f, 0.0f); glVertex3f(0.0f, 0.0f, 0.0f);
	glTexCoord2f(0.0f, 1.0f); glVertex3f(0.0f, 1.0f, 0.0f);
	glTexCoord2f(1.0f, 1.0f); glVertex3f(1.0f, 1.0f, 0.0f);
	glTexCoord2f(1.0f, 0.0f); glVertex3f(1.0f, 0.0f, 0.0f);
	glEnd();	

	glBindTexture(GL_TEXTURE_2D, reconstructedDataTexID);
	glBegin(GL_QUADS);
	glTexCoord2f(0.0f, 0.0f); glVertex3f(1.0f, 0.0f, 0.0f);
	glTexCoord2f(0.0f, 1.0f); glVertex3f(1.0f, 1.0f, 0.0f);
	glTexCoord2f(1.0f, 1.0f); glVertex3f(2.0f, 1.0f, 0.0f);
	glTexCoord2f(1.0f, 0.0f); glVertex3f(2.0f, 0.0f, 0.0f);
	glEnd();

	glEnable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
}


// render the data with marching squares, left side: original data, right side: reconstructed data
void RenderManager::renderMarchSquares(std::vector<float> dataleft, std::vector<float> dataright, double IsoLevel)
{
		int DimX = PCAComponent->getDimX();
		int DimY = PCAComponent->getDimY();

		glDisable(GL_TEXTURE_2D);
		glDisable(GL_DEPTH_TEST);
		glDisable(GL_LIGHTING);
		//set up the view parameters 
		glMatrixMode( GL_PROJECTION );
		glLoadIdentity();
		glOrtho(0,2*DimX,0,DimY,-1,1);
		glMatrixMode( GL_MODELVIEW );
		glLoadIdentity();
		
		glColor3d(1.0, 0.0, 0.0);
		glLineWidth(4.0);		

		// render the left side
		MarchSquares::renderMS(IsoLevel, dataleft, DimX, DimY);

		// render the right side
		glTranslated(DimX,0,0);		
		MarchSquares::renderMS(IsoLevel, dataright, DimX, DimY);

		glEnable(GL_TEXTURE_2D);
		glEnable(GL_DEPTH_TEST);
		glEnable(GL_LIGHTING);
}


void RenderManager::renderMarchSquares(int DataSetNr, double IsoLevel)
{
	// show the same data on both sides
	renderMarchSquares(PCAComponent->getDataSet(DataSetNr), PCAComponent->getReconstructedDataSet(DataSetNr),IsoLevel); 
}

// render 3D data with marching cubes algorithm 
void RenderManager::renderMarchCubes(std::vector<float> dataleft, std::vector<float> dataright, double IsoLevel)
{
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_LIGHTING);
	setMCEnvironment();
	glLineWidth(1.0);
	
	// build up new display lists if frame is dirty
	if (DirtyFlag)
	{
		int xdim = PCAComponent->getDimX();
		int ydim = PCAComponent->getDimY();
		int zdim = PCAComponent->getDimZ();
		
		glDeleteLists(MCDisplayListID1,1);
		MCDisplayListID1 = glGenLists(1);

		glDeleteLists(MCDisplayListID2,1);
		MCDisplayListID2 = glGenLists(1);

		// build display list for original data set
		glNewList(MCDisplayListID1, GL_COMPILE);
		glBegin(GL_TRIANGLES);
		for (int X = 0; X < xdim - 1; X++)
			for(int Y = 0; Y < ydim - 1; Y++)
				for(int Z = 0; Z < zdim - 1; Z++)
				{
					std::vector<float> temp;
					temp.resize(8);
					//acquire the scalar values for one cell
					temp[0] = dataleft[Z*xdim*ydim+Y*xdim+X];
					temp[1] = dataleft[Z*xdim*ydim+Y*xdim+X+1];
					temp[2] = dataleft[Z*xdim*ydim+(Y+1)*xdim+X+1];
					temp[3] = dataleft[Z*xdim*ydim+(Y+1)*xdim+X];
					temp[4] = dataleft[(Z+1)*xdim*ydim+Y*xdim+X];
					temp[5] = dataleft[(Z+1)*xdim*ydim+Y*xdim+X+1];
					temp[6] = dataleft[(Z+1)*xdim*ydim+(Y+1)*xdim+X+1];
					temp[7] = dataleft[(Z+1)*xdim*ydim+(Y+1)*xdim+X];
					
					MarchCubes::renderMC(X,Y,Z,1,temp,IsoLevel);
				}
		glEnd();
		glEndList();


		// build display list for reconstructed data set
		glNewList(MCDisplayListID2, GL_COMPILE);
		glBegin(GL_TRIANGLES);
		for (int X = 0; X < xdim - 1; X++)
			for(int Y = 0; Y < ydim - 1; Y++)
				for(int Z = 0; Z < zdim - 1; Z++)
				{
					std::vector<float> temp;
					temp.resize(8);
					//acquire the scalar values for one cell
					temp[0] = dataright[Z*xdim*ydim+Y*xdim+X];
					temp[1] = dataright[Z*xdim*ydim+Y*xdim+X+1];
					temp[2] = dataright[Z*xdim*ydim+(Y+1)*xdim+X+1];
					temp[3] = dataright[Z*xdim*ydim+(Y+1)*xdim+X];
					temp[4] = dataright[(Z+1)*xdim*ydim+Y*xdim+X];
					temp[5] = dataright[(Z+1)*xdim*ydim+Y*xdim+X+1];
					temp[6] = dataright[(Z+1)*xdim*ydim+(Y+1)*xdim+X+1];
					temp[7] = dataright[(Z+1)*xdim*ydim+(Y+1)*xdim+X];
					
					MarchCubes::renderMC(X,Y,Z,1,temp,IsoLevel);
				}
		glEnd();
		glEndList();

		DirtyFlag = false;
	}

	// set up camera transformation
	glMatrixMode( GL_PROJECTION );
	glLoadMatrixf( camera.getProjectionMatrix().ma );
	glMatrixMode( GL_MODELVIEW );
	glLoadMatrixf( (GLfloat *)camera.getViewMatrix().ma );
	
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glColor3d(0.0,0.0,0.0);
	
	int DimX = PCAComponent->getDimX();
	int DimY = PCAComponent->getDimY();
	int DimZ = PCAComponent->getDimZ();
	
	//render left side of screen
	glViewport(0,0,WindowWidth/2,WindowHeight);
	// render left volume boundaries
	glPushMatrix();
		glTranslatef(DimX/2,DimY/2,DimZ/2);
		glScalef(DimX,DimY,DimZ);
		glutWireCube(1);
	glPopMatrix();
	glCallList(MCDisplayListID1);

	//render right side of screen
	glViewport(WindowWidth/2,0,WindowWidth/2,WindowHeight);
	// render right volume boundaries
	glPushMatrix();
		glTranslatef(DimX/2,DimY/2,DimZ/2);
		glScalef(DimX,DimY,DimZ);
		glutWireCube(1);
	glPopMatrix();
	glCallList(MCDisplayListID2);

    // render separation line in the middle of the screen
	glViewport(0,0,WindowWidth,WindowHeight);	
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	glOrtho(0,2,0,1,-1,1);
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();	
	glBegin(GL_LINES);
		glVertex2d(1,1);
		glVertex2d(1,0);
	glEnd();
	    
	glDisable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);
}

// render marching cubes with given data set number
void RenderManager::renderMarchCubes(int DataSetNr, double IsoLevel)
{
	// get current original data set
	std::vector<float> data = PCAComponent->getDataSet(DataSetNr);
	// get current reconstructed data set
	std::vector<float> reconstructedData = PCAComponent->getReconstructedDataSet(DataSetNr);
	renderMarchCubes(data, reconstructedData,IsoLevel);

}

// render interpolated data sets with marching cubes, we need to adapt info overlay 
void RenderManager::renderMarchCubes(float DataSetNr,double IsoLevel, int showInfo, int PrincCompNr)
{
	std::vector<float> InterpolatedOrigData, InterpolatedReconstruction;

	int first = (int) DataSetNr;
	float t = DataSetNr - (float) first;
	
	InterpolatedOrigData = getPCAComponent()->getInterpolatedDataSet(first,first+1,t);
	InterpolatedReconstruction = getPCAComponent()->getInterpolatedReconstructedDataSet(first,first+1,t);

	renderMarchCubes(InterpolatedOrigData, InterpolatedReconstruction,IsoLevel);
	if(showInfo)
		showInfos(true, first, IsoLevel, PrincCompNr);

}

// visualize time evolution of one specific feature
void RenderManager::renderFeatureVector(int nr, int CurrentSet)
{			
		glDisable(GL_TEXTURE_2D);
		glDisable(GL_DEPTH_TEST);
		glDisable(GL_LIGHTING);
		glMatrixMode( GL_PROJECTION );
		glLoadIdentity();

		// setup window: x-scale = nr of sets, y-scale = max./min. value of corresponding feature
		int XScale = PCAComponent->getNrOfDataSets();
		int YMax = 0, YMin = 0;

		// Search for max./min. value
		for(int i=0; i < XScale; i++)
		{
			if (PCAComponent->getFeatureVector(nr)[i] > YMax)
				YMax = PCAComponent->getFeatureVector(nr)[i];

			if (PCAComponent->getFeatureVector(nr)[i] < YMin)
				YMin = PCAComponent->getFeatureVector(nr)[i];
		}

		glOrtho(0,XScale,YMin,YMax,-1,1);
		glMatrixMode( GL_MODELVIEW );
		glLoadIdentity();
		
		// visualize coordinate system
		glColor3d(0.5, 0.5, 0.5);
		glLineWidth(1.0);
		glBegin(GL_LINES);
			glVertex2d(0,0);
			glVertex2d(XScale,0);	
		glEnd();
		
		std::string s = "0";
		renderTextLine(0, 1, s);
				
		glColor3d(1.0, 0.0, 0.0);
		glLineWidth(3.0);

		
		glBegin(GL_LINE_STRIP); // draw the evolution function
		for(int i=0; i < XScale; i++)
		{
			glVertex2d(i,PCAComponent->getFeatureVector(nr)[i]);
		}
		glEnd();
		
		// visualize the position of the current data set with a vertical line
		glPointSize(10.0);
		glBegin(GL_POINTS);
			glVertex2d(CurrentSet, PCAComponent->getFeatureVector(nr)[CurrentSet]);
		glEnd();

		glColor3d(0.0, 0.0, 0.0);
		glLineWidth(1.0);
		glBegin(GL_LINES);
			glVertex2d(CurrentSet,YMin);
			glVertex2d(CurrentSet,YMax);		
		glEnd();

		// magnitude of current feature in current data set
		std::stringstream ss;
		ss << PCAComponent->getFeatureVector(nr)[CurrentSet];
		renderTextLine(CurrentSet + 1, 1, ss.str());

		glEnable(GL_TEXTURE_2D);
		glEnable(GL_DEPTH_TEST);
		glEnable(GL_LIGHTING);
		
}

// render interpolated data with textures
void RenderManager::renderInterpolatedData(float DataSetNr, int showTextures, int showMarchingSquaresOverlay, int showInfo, int showReconstructedData, double IsoLevel, int PrincCompNr, bool clamp)
{
	std::vector<float> InterpolatedOrigData, InterpolatedReconstruction;

	int first = (int) DataSetNr;
	float t = DataSetNr - (float) first;
	
	InterpolatedOrigData = getPCAComponent()->getInterpolatedDataSet(first,first+1,t);
	InterpolatedReconstruction = getPCAComponent()->getInterpolatedReconstructedDataSet(first,first+1,t);

	buildTextureFromDataVector(PCAComponent->getDimX(),PCAComponent->getDimY(), InterpolatedOrigData, InterpolatedReconstruction, clamp);
	if(showTextures)
		renderTextures();
	if(showMarchingSquaresOverlay)
		renderMarchSquares(InterpolatedOrigData, InterpolatedReconstruction, IsoLevel);
	if(showInfo)
		showInfos((bool) showReconstructedData, first, IsoLevel, PrincCompNr);

	glutSwapBuffers();
}

// render info overlay with given informations
void RenderManager::showInfos(bool showRecalculatedData, int DataSetNr, double IsoLevel, int princCompNr)
{
	
	int PaddingLeft=20, PaddingTop=20;
	
	glColor3d(0.0, 0.0, 0.0);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0.0, WindowWidth, 0.0, WindowHeight);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
		
	// infos at the left side of the view window
	std::stringstream StatsLeft[4];
	StatsLeft[0]<<"                   - ORIGINAL DATA -";
	StatsLeft[1]<<"<i> Toggle Info";
	StatsLeft[2]<<"Total count of data sets: "<<PCAComponent->getNrOfDataSets();
	StatsLeft[3]<<"Current data set: "<<DataSetNr<<" (use <LEFT/RIGHT> or mouse wheel to step)";
	
    
	for(unsigned i=0; i<4;i++)
	{
		renderTextLine(PaddingLeft,WindowHeight-PaddingTop-i*PaddingTop,StatsLeft[i].str());
	}

	// infos at the right side of the view window
	std::stringstream StatsRight;
	float CompressionRatio = (float)((int) (PCAComponent->getCompressionRatio()*100))/100;
	
	if(showRecalculatedData) // we can show informations about reconstructed data...
	{
		StatsRight<<"- RECONSTRUCTED DATA -     ("<<PCAComponent->getNrOfBasisStates()<<" PCs used for reconstruction, compression ratio: "<<CompressionRatio<<")";
	}
	else //or the number of the current PC
	{	
		StatsRight<<"- PRINCIPAL COMPONENT "<<princCompNr<<" -";
	}
	renderTextLine(PaddingLeft+WindowWidth/2,WindowHeight-PaddingTop,StatsRight.str());
	

	glMatrixMode(GL_MODELVIEW);
    glPopMatrix();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_LIGHTING);
	

}






